#------------------------------------------------------------------------------
# Name:         test_tools.py
# Purpose:      Test tools from nansat.tools
# Author:       Artem Moiseev
# Created:      17.01.2020
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#------------------------------------------------------------------------------

import os
import unittest
from mock import patch, DEFAULT

import numpy as np
try:
    import matplotlib.pyplot as plt
    plt.switch_backend('Agg')
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True
try:
    import cartopy
except ImportError:
    CARTOPY_IS_INSTALLED = False
else:
    CARTOPY_IS_INSTALLED = True

import nansat
from nansat.domain import Domain
from nansat.figure import Image
from nansat.nansat import Nansat as NANSAT
from nansat.tests import nansat_test_data as ntd
from nansat.tools import (distance2coast,
                            register_colormaps,
                            get_domain_map,
                            show_domain_map,
                            save_domain_map,
                            get_domain_map)


class ToolsTest(unittest.TestCase):
    def setUp(self):
        self.d = Domain(4326, "-te 25 70 35 72 -ts 100 100")
        # define a test Nansat object
        test_domain = Domain(4326, "-lle -180 -90 180 90 -ts 500 500")
        self.n = NANSAT.from_domain(test_domain, array=np.ones([500,500]))

    @patch('nansat.tools.os.getenv')
    def test_distance2coast_source_not_exists_envvar(self, mock_getenv):
        mock_getenv.return_value='/path/dos/not/exist'
        with self.assertRaises(IOError) as err:
            distance2coast(self.d)
        self.assertEqual('Distance to the nearest coast product does not exist - '
                         'see Nansat documentation to get it (the path is '
                         '/path/dos/not/exist)', str(err.exception))

    def test_distance2coast_source_not_exists_attribute(self):
        with self.assertRaises(IOError) as err:
            distance2coast(self.d, distance_src='/path/dos/not/exist')
        self.assertEqual('Distance to the nearest coast product does not exist - '
                         'see Nansat documentation to get it (the path is '
                         '/path/dos/not/exist)', str(err.exception))

    @patch.multiple('nansat.tools', Nansat=DEFAULT, os=DEFAULT)
    def test_distance2coast_integration(self, Nansat, os):
        Nansat.return_value = self.n
        os.path.exists.return_value=True
        result = distance2coast(self.d)
        self.assertEqual(type(result), NANSAT)

    def test_warning(self):
        register_colormaps()
        with self.assertWarns(UserWarning) as w:
            register_colormaps()

    @unittest.skipUnless(CARTOPY_IS_INSTALLED, 'Cartopy is required')
    def test_get_domain_map(self):
        ax = get_domain_map(self.d)
        self.assertIsInstance(ax, plt.Axes)

    def test_get_domain_map_no_cartopy(self):
        nansat.tools.CARTOPY_IS_INSTALLED = False
        with self.assertRaises(ImportError) as err:
            ax = get_domain_map(self.d)
        self.assertIn('Cartopy is not installed', str(err.exception))
        nansat.tools.CARTOPY_IS_INSTALLED = CARTOPY_IS_INSTALLED

    @unittest.skipUnless(CARTOPY_IS_INSTALLED, 'Cartopy is required')
    def test_save_domain_map(self):
        tmpfilename = os.path.join(ntd.tmp_data_path, 'domain_save_map.png')
        save_domain_map(self.d, tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))
        i = Image.open(tmpfilename)
        i.verify()
        # with cartopy>=0.20.0, the dpi attribute is not
        # always (100, 100)
        self.assertEqual(tuple(np.round(i.info['dpi'], 1)), (100, 100))



# Name:        mapper_masr2_l3
# Purpose:     Mapping for L3 data from the OBPG web-site
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from dateutil.parser import parse
import datetime
import os.path
import glob
import json

import pythesint as pti

import numpy as np

from nansat.vrt import VRT
from nansat.nsr import NSR
from nansat.utils import gdal, ogr
from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    ''' Mapper for Level-3 AMSR2 data from https://gcom-w1.jaxa.jp'''

    freqs = [6, 7, 10, 18, 23, 36, 89]

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' OBPG L3 VRT '''

        # test the product
        try:
            assert gdalMetadata['PlatformShortName'] == 'GCOM-W1'
            assert gdalMetadata['SensorShortName'] == 'AMSR2'
            assert gdalMetadata['ProductName'] == 'AMSR2-L3'
        except:
            raise WrongMapperError

        # get list of similar (same date, A/D orbit) files in the directory
        iDir, iFile = os.path.split(filename)
        iFileMask = iFile[:30] + '%02d' + iFile[32:]
        simFiles = []
        for freq in self.freqs:
            simFile = os.path.join(iDir, iFileMask % freq)
            #print simFile
            if os.path.exists(simFile):
                simFiles.append(simFile)

        metaDict = []
        for freq in self.freqs:
            simFile = os.path.join(iDir, iFileMask % freq)
            if simFile not in simFiles:
                continue
            #print 'simFile', simFile
            # open file, get metadata and get parameter name
            simSupDataset = gdal.Open(simFile)
            if simSupDataset is None:
                # skip this similar file
                #print 'No dataset: %s not a supported SMI file' % simFile
                continue
            # get subdatasets from the similar file
            simSubDatasets = simSupDataset.GetSubDatasets()
            for simSubDataset in simSubDatasets:
                #print 'simSubDataset', simSubDataset
                if 'Brightness_Temperature' in simSubDataset[0]:
                    # get SourceFilename from subdataset
                    metaEntry = {
                        'src': {'SourceFilename': simSubDataset[0],
                                'SourceBand': 1,
                                'ScaleRatio': 0.0099999998,
                                'ScaleOffset': 0},
                        'dst': {'wkv': 'brightness_temperature',
                                'frequency': '%02d' % freq,
                                'polarisation': simSubDataset[0][-2:-1],
                                'suffix': ('%02d%s' %
                                           (freq, simSubDataset[0][-2:-1]))}}
                    metaDict.append(metaEntry)

        # initiate VRT for the NSIDC 10 km grid
        self._init_from_dataset_params(760, 1120, (-3850000, 10000, 0.0, 5850000, 0.0, -10000), NSR(3411).wkt)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # Adding valid time to dataset
        self.dataset.SetMetadataItem('time_coverage_start', parse(gdalMetadata['ObservationStartDateTime']).isoformat())
        self.dataset.SetMetadataItem('time_coverage_end', parse(gdalMetadata['ObservationStartDateTime']).isoformat())

        mm = pti.get_gcmd_instrument('AMSR2')
        ee = pti.get_gcmd_platform('GCOM-W1')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))




# Name:         mapper_radarsat2
# Purpose:      Nansat mapping for Radarsat2 data
# Authors:      Morten W. Hansen, Knut-Frode Dagestad, Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from __future__ import unicode_literals, division, absolute_import
import os
import zipfile
import json
from dateutil.parser import parse

import numpy as np
try:
    import scipy.ndimage
except:
    IMPORT_SCIPY = False
else:
    IMPORT_SCIPY = True

import pythesint as pti

from nansat.nsr import NSR
from nansat.vrt import VRT
from nansat.domain import Domain
from nansat.node import Node
from nansat.utils import initial_bearing, gdal
from nansat.exceptions import WrongMapperError, NansatReadError


class Mapper(VRT):
    ''' Create VRT with mapping of WKV for Radarsat2 '''

    def __init__(self, inputFileName, gdalDataset, gdalMetadata,
                 xmlonly=False,  **kwargs):
        ''' Create Radarsat2 VRT '''
        fPathName, fExt = os.path.splitext(inputFileName)

        if zipfile.is_zipfile(inputFileName):
            # Open zip file using VSI
            fPath, fName = os.path.split(fPathName)
            filename = '/vsizip/%s/%s' % (inputFileName, fName)
            if not 'RS' in fName[0:2]:
                raise WrongMapperError('%s: Provided data is not Radarsat-2'
                                       % fName)
            gdalDataset = gdal.Open(filename)
            gdalMetadata = gdalDataset.GetMetadata()
        else:
            filename = inputFileName

        # if it is not RADARSAT-2, return
        if (not gdalMetadata or not 'SATELLITE_IDENTIFIER' in list(gdalMetadata.keys())):
            raise WrongMapperError(filename)
        elif gdalMetadata['SATELLITE_IDENTIFIER'] != 'RADARSAT-2':
            raise WrongMapperError(filename)

        if zipfile.is_zipfile(inputFileName):
            # Open product.xml to get additional metadata
            zz = zipfile.ZipFile(inputFileName)
            productXmlName = os.path.join(os.path.basename(
                inputFileName).split('.')[0], 'product.xml')
            productXml = zz.open(productXmlName).read()
        else:
            # product.xml to get additionali metadata
            productXmlName = os.path.join(filename, 'product.xml')
            if not os.path.isfile(productXmlName):
                raise WrongMapperError(filename)
            productXml = open(productXmlName).read()

        if not IMPORT_SCIPY:
            raise NansatReadError('Radarsat-2 data cannot be read because scipy is not installed')

        # parse product.XML
        rs2_0 = Node.create(productXml)

        if xmlonly:
            self.init_from_xml(rs2_0, filename)
            return

        # Get additional metadata from product.xml
        rs2_1 = rs2_0.node('sourceAttributes')
        rs2_2 = rs2_1.node('radarParameters')
        if rs2_2['antennaPointing'].lower() == 'right':
            antennaPointing = 90
        else:
            antennaPointing = -90
        rs2_3 = rs2_1.node('orbitAndAttitude').node('orbitInformation')
        passDirection = rs2_3['passDirection']

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset)
        self.dataset.SetGCPs(self.dataset.GetGCPs(), NSR().wkt)

        # define dictionary of metadata and band specific parameters
        pol = []
        metaDict = []

        # Get the subdataset with calibrated sigma0 only
        for dataset in gdalDataset.GetSubDatasets():
            if dataset[1] == 'Sigma Nought calibrated':
                s0dataset = gdal.Open(dataset[0])
                s0datasetName = dataset[0][:]
                band = s0dataset.GetRasterBand(1)
                s0datasetPol = band.GetMetadata()['POLARIMETRIC_INTERP']
                for i in range(1, s0dataset.RasterCount+1):
                    iBand = s0dataset.GetRasterBand(i)
                    polString = iBand.GetMetadata()['POLARIMETRIC_INTERP']
                    suffix = polString
                    # The nansat data will be complex
                    # if the SAR data is of type 10
                    dtype = iBand.DataType
                    if dtype == 10:
                        # add intensity band
                        metaDict.append(
                            {'src': {'SourceFilename':
                                     ('RADARSAT_2_CALIB:SIGMA0:'
                                      + filename + '/product.xml'),
                                     'SourceBand': i,
                                     'DataType': dtype},
                             'dst': {'wkv': 'surface_backwards_scattering_coefficient_of_radar_wave',
                                     'PixelFunctionType': 'intensity',
                                     'SourceTransferType': gdal.GetDataTypeName(dtype),
                                     'suffix': suffix,
                                     'polarization': polString,
                                     'dataType': 6}})
                        # modify suffix for adding the compled band below
                        suffix = polString+'_complex'
                    pol.append(polString)
                    metaDict.append(
                        {'src': {'SourceFilename': ('RADARSAT_2_CALIB:SIGMA0:'
                                                    + filename
                                                    + '/product.xml'),
                                 'SourceBand': i,
                                 'DataType': dtype},
                         'dst': {'wkv': 'surface_backwards_scattering_coefficient_of_radar_wave',
                                 'suffix': suffix,
                                 'polarization': polString}})

            if dataset[1] == 'Beta Nought calibrated':
                b0dataset = gdal.Open(dataset[0])
                b0datasetName = dataset[0][:]
                for j in range(1, b0dataset.RasterCount+1):
                    jBand = b0dataset.GetRasterBand(j)
                    polString = jBand.GetMetadata()['POLARIMETRIC_INTERP']
                    if polString == s0datasetPol:
                        b0datasetBand = j

        ###############################
        # Add SAR look direction
        ###############################
        d = Domain(ds=gdalDataset)
        lon, lat = d.get_geolocation_grids(100)

        '''
        (GDAL?) Radarsat-2 data is stored with maximum latitude at first
        element of each column and minimum longitude at first element of each
        row (e.g. np.shape(lat)=(59,55) -> latitude maxima are at lat[0,:],
        and longitude minima are at lon[:,0])

        In addition, there is an interpolation error for direct estimate along
        azimuth. We therefore estimate the heading along range and add 90
        degrees to get the "satellite" heading.

        '''
        if str(passDirection).upper() == 'DESCENDING':
            sat_heading = initial_bearing(lon[:, :-1], lat[:, :-1],
                                          lon[:, 1:], lat[:, 1:]) + 90
        elif str(passDirection).upper() == 'ASCENDING':
            sat_heading = initial_bearing(lon[:, 1:], lat[:, 1:],
                                          lon[:, :-1], lat[:, :-1]) + 90
        else:
            print('Can not decode pass direction: ' + str(passDirection))

        # Calculate SAR look direction
        look_direction = sat_heading + antennaPointing
        # Interpolate to regain lost row
        look_direction = np.mod(look_direction, 360)
        look_direction = scipy.ndimage.interpolation.zoom(
            look_direction, (1, 11./10.))
        # Decompose, to avoid interpolation errors around 0 <-> 360
        look_direction_u = np.sin(np.deg2rad(look_direction))
        look_direction_v = np.cos(np.deg2rad(look_direction))
        look_u_VRT = VRT.from_array(look_direction_u)
        look_v_VRT = VRT.from_array(look_direction_v)

        # Note: If incidence angle and look direction are stored in
        #       same VRT, access time is about twice as large
        lookVRT = VRT.from_lonlat(lon, lat)
        lookVRT.create_band(
            [{'SourceFilename': look_u_VRT.filename, 'SourceBand': 1},
             {'SourceFilename': look_v_VRT.filename, 'SourceBand': 1}],
            {'PixelFunctionType': 'UVToDirectionTo'})

        # Blow up to full size
        lookVRT = lookVRT.get_resized_vrt(gdalDataset.RasterXSize, gdalDataset.RasterYSize)
        # Store VRTs so that they are accessible later
        self.band_vrts['look_u_VRT'] = look_u_VRT
        self.band_vrts['look_v_VRT'] = look_v_VRT
        self.band_vrts['lookVRT'] = lookVRT

        # Add band to full sized VRT
        lookFileName = self.band_vrts['lookVRT'].filename
        metaDict.append({'src': {'SourceFilename': lookFileName,
                                 'SourceBand': 1},
                         'dst': {'wkv': 'sensor_azimuth_angle',
                                 'name': 'look_direction'}})

        ###############################
        # Create bands
        ###############################
        self.create_bands(metaDict)

        ###################################################
        # Add derived band (incidence angle) calculated
        # using pixel function "BetaSigmaToIncidence":
        ###################################################
        src = [{'SourceFilename': b0datasetName,
                'SourceBand':  b0datasetBand,
                'DataType': dtype},
               {'SourceFilename': s0datasetName,
                'SourceBand': 1,
                'DataType': dtype}]
        dst = {'wkv': 'angle_of_incidence',
               'PixelFunctionType': 'BetaSigmaToIncidence',
               'SourceTransferType': gdal.GetDataTypeName(dtype),
               '_FillValue': -10000,   # NB: this is also hard-coded in
                                       #     pixelfunctions.c
               'dataType': 6,
               'name': 'incidence_angle'}

        self.create_band(src, dst)
        self.dataset.FlushCache()

        ###################################################################
        # Add sigma0_VV - pixel function of sigma0_HH and beta0_HH
        # incidence angle is calculated within pixel function
        # It is assummed that HH is the first band in sigma0 and
        # beta0 sub datasets
        ###################################################################
        if 'VV' not in pol and 'HH' in pol:
            s0datasetNameHH = pol.index('HH')+1
            src = [{'SourceFilename': s0datasetName,
                    'SourceBand': s0datasetNameHH,
                    'DataType': 6},
                   {'SourceFilename': b0datasetName,
                    'SourceBand': b0datasetBand,
                    'DataType': 6}]
            dst = {'wkv': 'surface_backwards_scattering_coefficient_of_radar_wave',
                   'PixelFunctionType': 'Sigma0HHBetaToSigma0VV',
                   'polarization': 'VV',
                   'suffix': 'VV'}
            self.create_band(src, dst)
            self.dataset.FlushCache()

        ############################################
        # Add SAR metadata
        ############################################
        if antennaPointing == 90:
            self.dataset.SetMetadataItem('ANTENNA_POINTING', 'RIGHT')
        if antennaPointing == -90:
            self.dataset.SetMetadataItem('ANTENNA_POINTING', 'LEFT')
        self.dataset.SetMetadataItem('ORBIT_DIRECTION',
                                     str(passDirection).upper())

        # set valid time
        self.dataset.SetMetadataItem('time_coverage_start',
                                     (parse(gdalMetadata['FIRST_LINE_TIME']).
                                      isoformat()))
        self.dataset.SetMetadataItem('time_coverage_end',
                                     (parse(gdalMetadata['LAST_LINE_TIME']).
                                      isoformat()))

        # Get dictionary describing the instrument and platform according to
        # the GCMD keywords
        mm = pti.get_gcmd_instrument("C-SAR")
        ee = pti.get_gcmd_platform('radarsat-2')

        # TODO: Validate that the found instrument and platform are indeed what we
        # want....

        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))
        self.dataset.SetMetadataItem('entry_title', 'Radarsat-2 SAR')
        self.dataset.SetMetadataItem('provider', 'MDA/GSI')
        self.dataset.SetMetadataItem('dataset_parameters', json.dumps(
                                     ['surface_backwards_scattering_coefficient_of_radar_wave']))
        self.dataset.SetMetadataItem('entry_id', os.path.basename(filename))

    def init_from_xml(self, productXml, filename):
        ''' Fast init from metada in XML only '''
        numberOfLines = int(productXml
                            .node('imageAttributes')
                            .node('rasterAttributes')
                            .node('numberOfLines')
                            .value)
        numberOfSamples = int(productXml
                              .node('imageAttributes')
                              .node('rasterAttributes')
                              .node('numberOfSamplesPerLine')
                              .value)

        VRT.__init__(self, srcRasterXSize=numberOfSamples, srcRasterYSize=numberOfLines)

        gcps = []
        geogrid = productXml.node(
            'imageAttributes').node('geographicInformation').node('geolocationGrid')
        for child in geogrid.children:
            pix = float(child.node('imageCoordinate').node('pixel').value)
            lin = float(child.node('imageCoordinate').node('line').value)
            lon = float(child.node('geodeticCoordinate').node('longitude').value)
            lat = float(child.node('geodeticCoordinate').node('latitude').value)
            gcps.append(gdal.GCP(lon, lat, 0, pix, lin))

        self.dataset.SetGCPs(gcps, NSR().wkt)

        dates = list(map(parse, [child.node('timeStamp').value for child in
                                 (productXml.node('sourceAttributes')
                                  .node('orbitAndAttitude')
                                  .node('orbitInformation')
                                  .nodeList('stateVector'))]))

        self.dataset.SetMetadataItem('time_coverage_start', min(dates).isoformat())
        self.dataset.SetMetadataItem('time_coverage_end', max(dates).isoformat())
        self.dataset.SetMetadataItem('platform', json.dumps(pti.get_gcmd_platform('radarsat-2')))
        self.dataset.SetMetadataItem('instrument', json.dumps(pti.get_gcmd_instrument('C-SAR')))
        self.dataset.SetMetadataItem('Entry Title', 'Radarsat-2 SAR')
        self.dataset.SetMetadataItem('Data Center', 'CSA')
        self.dataset.SetMetadataItem('ISO Topic Category', 'Oceans')
        self.dataset.SetMetadataItem('Summary', 'Radarsat-2 SAR data')
        self.dataset.SetMetadataItem('provider', 'MDA/GSI')
        self.dataset.SetMetadataItem('dataset_parameters', json.dumps(
            'surface_backwards_scattering_coefficient_of_radar_wave'))
        self.dataset.SetMetadataItem('entry_id', os.path.basename(filename))



# Name:         mapper_metno_hires_seaice.py
# Purpose:      Nansat mapping for high resolution sea ice
#                   from met.no Thredds server
# Authors:      Knut-Frode Dagestad
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html

# High resolution (1 km) manual ice concentration, based on SAR imagery
# http://thredds.met.no/thredds/catalog/myocean/siw-tac/siw-metno-svalbard/
#
# Mapper may be called with full URL:
#   'http://thredds.met.no/thredds/dodsC/myocean/siw-tac/siw-metno-svalbard/2014/01/ice_conc_svalbard_201401091500.nc
# or with keyword (fake filename):
#   'metno_hires_seaice:20140109'
#
# The latter syntax will construct the URL,
# and will return the closest available data within +/- 3 days
import sys
try:
    import urllib3 as urllib
except:
    import urllib2 as urllib

from datetime import datetime, timedelta

from nansat.utils import gdal, ogr, osr
from nansat.exceptions import WrongMapperError
from nansat.vrt import VRT


class Mapper(VRT):
    ''' Create VRT with mapping of WKV for Met.no seaice '''

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' Create VRT '''

        ThreddsBase = 'http://thredds.met.no/thredds/dodsC/myocean/siw-tac/siw-metno-svalbard/'
        # First check if mapper is called with keyword syntax:
        # filename = metno_hires_seaice:YYYYmmdd
        keywordBase = 'metno_hires_seaice'
        foundDataset = False
        if filename[0:len(keywordBase)] == keywordBase:
            keywordTime = filename[len(keywordBase)+1:]
            requestedTime = datetime.strptime(keywordTime, '%Y%m%d')
            # Search for nearest available file, within the closest 3 days
            for deltaDay in [0, -1, 1, -2, 2, -3, 3]:
                validTime = (requestedTime + timedelta(days=deltaDay) +
                             timedelta(hours=15))
                filename = (ThreddsBase +
                            validTime.strftime(
                                '%Y/%m/ice_conc_svalbard_%Y%m%d1500.nc'))
                try:
                    urllib.urlopen(filename + '.dds')
                    foundDataset = True
                    # Data is found for this day
                    break
                except:
                    # No data for this day
                    pass

        if not foundDataset:
            raise WrongMapperError

        # Then check if a valid OPeNDAP URL is given
        # (or has been constructed from keyword)
        if filename[0:len(ThreddsBase)] != ThreddsBase:
            AttributeError("Not Met.no Svalbard-ice Thredds URL")
        else:
            timestr = filename[-15:-3]
            validTime = datetime.strptime(timestr, '%Y%m%d%H%M')

        filename = filename + '?ice_concentration[0][y][x]'
        srcProjection = osr.SpatialReference()
        srcProjection.ImportFromProj4('+proj=stere lon_0=0.0 +lat_0=90 +datum=WGS84 +ellps=WGS84 +units=km +no_defs')
        srcProjection = srcProjection.ExportToWkt()

        # From thredds web, with manual shift
        srcGeotransform = (-1243.008 - 1, 1, 0, -3190.026 - 7, 0, 1)

        # create empty VRT dataset with geolocation only
        self._init_from_dataset_params(3812, 2980, srcGeotransform, srcProjection)

        metaDict = [{'src': {'SourceFilename': filename,
                             'sourceBand': 1},
                     'dst': {'name': 'sea_ice_area_fraction',
                             'wkv': 'sea_ice_area_fraction'}}]

        # Add band
        self.create_bands(metaDict)

        # Set time
        self.logger.info('Valid time: %s', str(validTime))
        self.dataset.SetMetadataItem('time_coverage_start',
                                     validTime.isoformat())



# Name:         exceptions.py
# Purpose:      Definitions of nansat exceptions
# Authors:      Morten W. Hansen
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from __future__ import absolute_import

class NansatProjectionError(Exception):
    """ Cannot get the projection """
    pass


class NansatGDALError(Exception):
    """ Error from GDAL  """
    pass


class NansatReadError(Exception):
    """ Exception if a file cannot be read with Nansat """
    pass

class NansatGeolocationError(Exception):
    """ Exception if geolocation is wrong (e.g., all lat/lon values are 0) """
    pass

class NansatMissingProjectionError(Exception):
    """ Exception raised if no (sub-) dataset has projection """

class WrongMapperError(Exception):
    """ Error for handling data that does not fit a given mapper """
    pass




# Name:         mapper_ncep_wind_online.py
# Purpose:      Nansat mapping for NCEP GFS model data, stored online
# Author:       Knut-Frode Dagestad, Morten W. Hansen
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#
# Mapper searches two online archives for NCEP GFS grib files
# covering the requested time, and downloads using curl, if found:
#    1. ftp://ftp.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/ (~last month)
#    2. http://nomads.ncdc.noaa.gov/data/gfs4/ (back to June 2012, with holes)
#
# Usage:
#    w = Nansat('ncep_wind_online:YYYYMMDDHHMM')
# Example:
#    w = Nansat('ncep_wind_online:201405011000')

from __future__ import absolute_import, print_function, division, unicode_literals

import os
import sys
from datetime import datetime, timedelta

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError
from nansat.nansat import Nansat

# Place to store downloads - this can be changed via the "outFolder" argument
# to Mapper.__init__
downloads = os.path.join(os.path.expanduser('~'), 'ncep_gfs_downloads')


class Mapper(VRT, object):
    """VRT with mapping of WKV for NCEP GFS"""

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 outFolder=downloads, **kwargs):
        """Create NCEP VRT"""

        if not os.path.exists(outFolder):
            os.mkdir(outFolder)

        ##############
        # Get time
        ##############
        keyword_base = 'ncep_wind_online'
        if filename[0:len(keyword_base)] != keyword_base:
            raise WrongMapperError

        time_str = filename[len(keyword_base)+1::]
        time = datetime.strptime(time_str, '%Y%m%d%H%M')
        print(time)

        ########################################
        # Find and download online grib file
        ########################################
        # Find closest 6 hourly modelrun and forecast hour
        model_run_hour = round((time.hour + time.minute/60.)/6)*6
        nearest_model_run = (datetime(time.year, time.month, time.day)
                             + timedelta(hours=model_run_hour))
        if sys.version_info < (2, 7):
            td = (time - nearest_model_run)
            forecast_hour = (td.microseconds +
                             (td.seconds + td.days * 24 * 3600)
                             * 10 ** 6) / 10 ** 6 / 3600.
        else:
            forecast_hour = (time - nearest_model_run).total_seconds() / 3600.
        if model_run_hour == 24:
            model_run_hour = 0
        if forecast_hour < 1.5:
            forecast_hour = 0
        else:
            forecast_hour = 3

        #########################################################
        # Try first to get NRT data from
        # ftp://ftp.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/
        # - avaliable approximately the latest month
        #########################################################
        url = ('ftp://ftp.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/' + 'gfs.'
               + nearest_model_run.strftime('%Y%m%d') + '%.2d' % model_run_hour + '/gfs.t' + '%.2d'
               % model_run_hour + 'z.master.grbf' +  '%.3d' % forecast_hour + '.10m.uv.grib2')
        out_filename = os.path.join(outFolder,
                                   ('ncep_gfs_' +
                                    nearest_model_run.strftime('%Y%m%d_%HH_') +
                                    '%.2d' % forecast_hour +
                                    '.10m.uv.grib2'))
        if os.path.exists(out_filename):
            print('NCEP wind is already downloaded: ' + out_filename)
        else:
            os.system('curl -so ' + out_filename + ' ' + url)
            if os.path.exists(out_filename):
                print('Downloaded ' + out_filename)
            else:
                print('NRT GRIB file not available: ' + url)
                #########################################################
                # If NRT file not available, search in long term archive
                #########################################################
                url = ('http://nomads.ncdc.noaa.gov/data/gfs4/' +
                       nearest_model_run.strftime('%Y%m/%Y%m%d/'))
                basename = ('gfs_4_' + nearest_model_run.strftime('%Y%m%d_') +
                            nearest_model_run.strftime('%H%M_') +
                            '%.3d' % forecast_hour)
                filename = basename + '.grb2'
                out_filename = os.path.join(outFolder, filename)
                print('Downloading ' + url + filename)

                # Download subset of grib file
                mapper_dir = os.path.dirname(os.path.abspath(__file__))
                get_inv = os.path.join(mapper_dir, 'get_inv.pl')
                if not os.path.isfile(get_inv):
                    raise IOError('%s: File not found' % get_inv)
                get_grib = os.path.join(mapper_dir, 'get_grib.pl')

                if not os.path.isfile(get_grib):
                    raise IOError('%s: File not found' % get_grib)

                if not os.path.isfile(out_filename):
                    command = (get_inv + ' ' + url + basename +
                               '.inv | egrep "(:UGRD:10 m |:VGRD:10 m )" | ' +
                               get_grib + ' ' + url + filename + ' ' + out_filename)
                    os.system(command)
                    if os.path.isfile(out_filename):
                        print('Downloaded ' + filename + ' to ' + outFolder)
                else:
                    print('Already downloaded %s' % out_filename)

                if not os.path.isfile(out_filename):
                    sys.exit('No NCEP wind files found for requested time')

        ######################################################
        # Open downloaded grib file with a(ny) Nansat mapper
        ######################################################
        w = Nansat(out_filename)
        self._copy_from_dataset(w.vrt.dataset)

        return

    #def __del__(self):
    #    super(Mapper,self).__del__()
    #    # Delete the downloaded ncep data
    #    try:
    #        os.unlink(self.dataset.GetFileList()[0])
    #    except:
    #        pass



# Name:        mapper_modisL1
# Purpose:     Mapping for MODIS-L1 data
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from dateutil.parser import parse
import warnings

from nansat.utils import gdal, ogr
from nansat.exceptions import WrongMapperError
from nansat.vrt import VRT


class HDF4Mapper(VRT):

    def find_metadata(self, iMetadata, iKey, default=''):
        """ Find metadata which has similar key

        Parameters
        ----------
            iMetadata : dict
                input metadata, usually gdalMetadata
            iKey : str
                key to search for
            default : str
                default value

        """
        value = default
        for key in iMetadata:
            if iKey in key:
                value = iMetadata[key]
                break

        return value



# ------------------------------------------------------------------------------
# Name:         test_nansat.py
# Purpose:      Test the Nansat class
#
# Author:       Morten Wergeland Hansen, Asuka Yamakawa, Anton Korosov
#
# Created:      18.06.2014
# Last modified:24.08.2017 14:00
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
# ------------------------------------------------------------------------------
from __future__ import print_function, absolute_import, division
import os
import sys
import json
import unittest
import warnings
import datetime
from dateutil.parser import parse
import tempfile
from xml.sax.saxutils import unescape
from mock import patch, PropertyMock, Mock, MagicMock, DEFAULT

import numpy as np

try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    import matplotlib
    import matplotlib.pyplot as plt
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True

from netCDF4 import Dataset

from nansat import Nansat, Domain, NSR
from nansat.utils import gdal
from nansat.tests.nansat_test_base import NansatTestBase

warnings.simplefilter("always", UserWarning)


class ExporterTest(NansatTestBase):

    def test_geolocation_of_exportedNC_vs_original(self):
        """ Lon/lat in original and exported file should coincide """
        orig = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        orig.export(self.tmp_filename)

        copy = Nansat(self.tmp_filename, mapper=self.default_mapper)
        lon0, lat0 = orig.get_geolocation_grids()
        lon1, lat1 = copy.get_geolocation_grids()
        np.testing.assert_allclose(lon0, lon1)
        np.testing.assert_allclose(lat0, lat1)

    def test_special_characters_in_exported_metadata(self):
        orig = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        orig.vrt.dataset.SetMetadataItem('jsonstring', json.dumps({'meta1':
                                         'hei', 'meta2': 'derr'}))
        orig.export(self.tmp_filename)
        copy = Nansat(self.tmp_filename, mapper=self.default_mapper)
        dd = json.loads(unescape(copy.get_metadata('jsonstring'), {'&quot;':
                                                                   '"'}))
        self.assertIsInstance(dd, dict)

    def test_time_coverage_metadata_of_exported_equals_original(self):
        orig = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        orig.set_metadata('time_coverage_start', '2010-01-02T08:49:02.347809')
        orig.set_metadata('time_coverage_end', '2010-01-02T08:50:03.599373')
        orig.export(self.tmp_filename)
        copy = Nansat(self.tmp_filename, mapper=self.default_mapper)

        self.assertEqual(orig.get_metadata('time_coverage_start'),
                copy.get_metadata('time_coverage_start'))
        self.assertEqual(orig.get_metadata('time_coverage_end'),
                copy.get_metadata('time_coverage_end'))

    def test_export_netcdf(self):
        """ Test export and following import of data with bands containing np.nan values """
        n = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        arrNoNaN = np.random.randn(n.shape()[0], n.shape()[1])
        n.add_band(arrNoNaN, {'name': 'testBandNoNaN'})
        arrWithNaN = arrNoNaN.copy()
        arrWithNaN[int(n.shape()[0] / 2.) - 10:int(n.shape()[0] / 2 + 10),
                   int(n.shape()[1] / 2.) - 10:int(n.shape()[1] / 2 + 10)] = np.nan
        n.add_band(arrWithNaN, {'name': 'testBandWithNaN'})
        n.export(self.tmp_filename)
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        earrNoNaN = exported['testBandNoNaN']
        # Use allclose to allow some roundoff errors
        self.assertTrue(np.allclose(arrNoNaN, earrNoNaN))
        earrWithNaN = exported['testBandWithNaN']
        np.testing.assert_allclose(arrWithNaN, earrWithNaN)

    def test_export_gcps_to_netcdf(self):
        """ Should export file with GCPs and write correct bands"""
        n0 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_export_gcps.nc')
        n0.export(tmpfilename)

        ncf = Dataset(tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))
        self.assertTrue('GCPX' in ncf.variables)
        self.assertTrue('GCPY' in ncf.variables)
        self.assertTrue('GCPPixel' in ncf.variables)
        self.assertTrue('GCPLine' in ncf.variables)

        n1 = Nansat(tmpfilename, mapper=self.default_mapper)
        b0 = n0['L_469']
        b1 = n1['L_469']
        np.testing.assert_allclose(b0, b1)

        lon0, lat0 = n0.get_geolocation_grids()
        lon1, lat1 = n1.get_geolocation_grids()
        np.testing.assert_allclose(lon0, lon1)
        np.testing.assert_allclose(lat0, lat1)

    def test_export_gcps_complex_to_netcdf(self):
        """ Should export file with GCPs and write correct complex bands"""
        n0 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        b0 = n0['L_469']

        n1 = Nansat.from_domain(n0)
        n1.add_band(b0.astype('complex64'), parameters={'name': 'L_469'})

        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_export_gcps_complex.nc')
        n1.export(tmpfilename)

        ncf = Dataset(tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))
        self.assertTrue('GCPX' in ncf.variables)
        self.assertTrue('GCPY' in ncf.variables)
        self.assertTrue('GCPPixel' in ncf.variables)
        self.assertTrue('GCPLine' in ncf.variables)

        n2 = Nansat(tmpfilename, mapper=self.default_mapper)
        b2 = n2['L_469']

        lon0, lat0 = n0.get_geolocation_grids()
        lon2, lat2 = n1.get_geolocation_grids()
        np.testing.assert_allclose(lon0, lon2)
        np.testing.assert_allclose(lat0, lat2)

    def test_export_gtiff(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_export.tif')
        n.export(tmpfilename, driver='GTiff')

        self.assertTrue(os.path.exists(tmpfilename))

    def test_export_band(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export_band.tif')
        n.export(tmpfilename, bands=[1], driver='GTiff')
        n = Nansat(tmpfilename, mapper=self.default_mapper)

        self.assertTrue(os.path.exists(tmpfilename))
        self.assertEqual(n.vrt.dataset.RasterCount, 1)

    def test_export_band_by_name(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export_band.tif')
        n.export(tmpfilename, bands=['L_645'], driver='GTiff')
        n = Nansat(tmpfilename, mapper=self.default_mapper)

        self.assertTrue(os.path.exists(tmpfilename))
        self.assertEqual(n.vrt.dataset.RasterCount, 1)

    def test_reproject_and_export_band(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n2 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n1.reproject(n2)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_export_band.nc')
        n1.export(tmpfilename, bands=[1])

        n = Nansat(tmpfilename, mapper=self.default_mapper)
        self.assertTrue(os.path.exists(tmpfilename))
        self.assertEqual(n.vrt.dataset.RasterCount, 1)

    def test_export_selected_bands(self):
        n = Nansat(self.test_file_gcps, mapper=self.default_mapper)
        resfile = 'tmp.nc'
        new_band = np.random.randn(n.shape()[0], n.shape()[1])
        n.add_band(new_band, {'name': 'newBand'})
        # Test with band numbers
        n.export(resfile, bands=[4, 2])
        self.assertTrue(os.path.exists(resfile))
        nn = Nansat(resfile, mapper=self.default_mapper)
        self.assertTrue(nn.has_band('newBand'))
        self.assertTrue(nn.has_band('L_555'))
        os.unlink(resfile)
        # Test with band names - not yet implemented
#         n.export(resfile, bands=['newBand', 'L_555'])
#         nn = Nansat(resfile)
#         self.assertTrue(nn.has_band('newBand'))
#         self.assertTrue(nn.has_band('L_555'))
#         os.unlink(resfile)

    def test_export_option(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export_option.nc')
        # Test with band numbers
        n.export(tmpfilename, options='WRITE_LONLAT=YES')
        n.export(tmpfilename + '2', options=['WRITE_LONLAT=NO'])
        nn = Nansat(tmpfilename, mapper=self.default_mapper)
        nn2 = Nansat(tmpfilename + '2', mapper=self.default_mapper)
        self.assertTrue(nn.has_band('lon'))
        self.assertTrue(nn.has_band('lat'))
        self.assertTrue(nn.has_band('Bristol'))
        self.assertFalse(nn2.has_band('lon'))
        self.assertFalse(nn2.has_band('lat'))
        self.assertTrue(nn2.has_band('Bristol'))

    def test_export2thredds_arctic_long_lat(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper, log_level=40)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export2thredds_arctic.nc')
        bands = {
            'Bristol': {'type': '>i2'},
            'Bootstrap': {'type': '>i2'},
            'UMass_AES': {'type': '>i2'},
        }
        n.export2thredds(tmpfilename, bands, time=datetime.datetime(2016, 1, 20))

        self.assertTrue(os.path.exists(tmpfilename))
        g = gdal.Open(tmpfilename)
        metadata = g.GetMetadata_Dict()

        # GDAL behaves differently:
        # Windows: nc-attributes are accessible without 'NC_GLOBAL#' prefix
        # Linux: nc-attributes are accessible only with 'NC_GLOBAL#' prefix
        # OSX: ?
        # Therefore we have to add NC_GLOBAL# and test if such metadata exists
        nc_prefix = 'NC_GLOBAL#'
        if not nc_prefix + 'easternmost_longitude' in metadata:
            nc_prefix = ''
        self.assertIn(nc_prefix + 'easternmost_longitude', metadata)

        # Test that the long/lat values are set correctly
        test_metadata_keys = ['easternmost_longitude', 'westernmost_longitude',
                              'northernmost_latitude', 'southernmost_latitude']
        test_metadata_min = [179, -180, 89.9, 53]
        test_metadata_max = [180, -179, 90, 54]
        for i, test_metadata_key in enumerate(test_metadata_keys):
            medata_value = float(metadata[nc_prefix + test_metadata_key])
            self.assertTrue(medata_value >= test_metadata_min[i],
                            '%s is wrong: %f'%(test_metadata_key, medata_value))
            self.assertTrue(medata_value <= test_metadata_max[i],
                            '%s is wrong: %f'%(test_metadata_key, medata_value))

    def test_dont_export2thredds_gcps(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n2 = Nansat.from_domain(n)
        n.add_band(np.ones(n2.shape(), np.float32))
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export2thredds.nc')
        with self.assertRaises(ValueError) as e:
            n2.export2thredds(tmpfilename)
        self.assertIn('Cannot export dataset with GCPS', e.exception.args[0])

    def test_export2thredds_longlat_list(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        with self.assertRaises(ValueError):
            n.export2thredds('aa', ['L_469'])

    def test_export2thredds_longlat_dict(self):
        d = Domain("+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs",
                   "-te 27 70 31 72 -ts 200 200")
        n = Nansat.from_domain(d)
        n.add_band(np.ones(d.shape(), np.float32),
                   parameters={'name': 'L_469'})
        n.set_metadata('time_coverage_start', '2016-01-19')

        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_export2thredds_longlat.nc')
        n.export2thredds(tmpfilename, {'L_469': {'type': '>i1'}})
        ncI = Dataset(tmpfilename, 'r')
        ncIVar = ncI.variables['L_469']
        self.assertTrue(ncIVar.grid_mapping in ncI.variables.keys())
        self.assertEqual(ncIVar[:].dtype, np.int8)

    def test_export_netcdf_complex_remove_meta(self):
        n = Nansat(self.test_file_complex, mapper=self.default_mapper)
        self.assertEqual(n.get_metadata('PRODUCT_TYPE'), 'SLC')
        n.export(self.tmp_filename, rm_metadata=['PRODUCT_TYPE'])
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        with self.assertRaises(ValueError):
            exported.get_metadata('PRODUCT_TYPE')
        self.assertTrue((n[1] == exported[1]).any())

    def test_export_netcdf_arctic(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        n.export(self.tmp_filename)
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        self.assertTrue((n[1] == exported[1]).any())
        self.assertTrue((n[2] == exported[2]).any())
        self.assertTrue((n[3] == exported[3]).any())

    def test_export_netcdf_arctic_hardcopy(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        n.export(self.tmp_filename, hardcopy=True)
        exported = Nansat(self.tmp_filename, mapper=self.default_mapper)
        self.assertTrue((n[1] == exported[1]).any())
        self.assertTrue((n[2] == exported[2]).any())
        self.assertTrue((n[3] == exported[3]).any())

    @patch('nansat.exporter.VRT._add_geolocation')
    def test_export_add_geoloc(self, mock_add_geolocation):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper)
        n.export(self.tmp_filename, add_geolocation=True)
        self.assertTrue(mock_add_geolocation.called)

    def test_export2thredds_rmmetadata(self):
        n = Nansat(self.test_file_arctic, mapper=self.default_mapper, log_level=40)
        n.export2thredds(self.tmp_filename, {'Bristol': {'type': '>i2'}},
                        time=datetime.datetime(2016, 1, 20),
                        rm_metadata=['description'])


#class TestExporter__export2thredds(unittest.TestCase):
class TestExporter__export2thredds(NansatTestBase):

    def setUp(self):
        super(TestExporter__export2thredds, self).setUp()
        fd, self.tmp_ncfile = tempfile.mkstemp(suffix='.nc')
        ds = Dataset(self.tmp_ncfile, 'w')
        xsz = 739
        ysz = 949
        timesz = 3

        # Set global metadat
        ds.setncattr('Conventions', 'CF-1.6')
        ds.setncattr('time_coverage_start', '2019-06-15T15:00:00.000000')
        ds.setncattr('time_coverage_end', '2019-06-15T21:00:00.000000')

        # Set dimensions
        ds.createDimension('y', ysz)
        ds.createDimension('x', xsz)
        ds.createDimension('time', timesz)
        # Set variables
        # 1d "dimensional" variables i.e lats, times, etc.
        times = ds.createVariable('time', 'i4', ('time'))
        times.units = 'seconds since 1970-01-01 00:00'
        times.standard_name = 'time'
        times.long_name = 'time'
        times[0] = np.ma.MaskedArray(data = 1560610800.0, mask =
                False, fill_value = 1e+20)
        times[1] = np.ma.MaskedArray(data = 1560621600.0, mask =
                False, fill_value = 1e+20)
        times[2] = np.ma.MaskedArray(data = 1560632400.0, mask =
                False, fill_value = 1e+20)

        xs = ds.createVariable('x', 'i4', ('x'))
        xs[:] = np.linspace(278603.16, 2123603.2, xsz)

        ys = ds.createVariable('y', 'i4', ('y'))
        ys[:] = np.linspace(-897931.56, 1472068.4, ysz)

        # Spatial variables 2d, 3d, and 4d
        xwind = ds.createVariable('x_wind_10m', 'f4', ('time', 'y', 'x'))
        xwind.standard_name = 'x_wind'
        xwind[:,:,:] = np.ones((timesz,ysz,xsz))*9.
        ywind = ds.createVariable('y_wind_10m', 'f4', ('time', 'y', 'x'))
        ywind.standard_name = 'y_wind'
        ywind[:,:,:] = np.ones((timesz,ysz,xsz))*10.
        mask = ds.createVariable('mask', 'i4', ('time', 'y', 'x'))
        mask.standard_name = 'mask'
        mask[:,:,:] = np.zeros((timesz,ysz,xsz))
        mask[:,:int(ysz/2),:int(xsz/2)] = 1

        # Projection
        proj = ds.createVariable('projection_lambert', 'f4', ())
        proj.grid_mapping_name = 'lambert_conformal_conic'
        proj.standard_parallel = np.array([ 77.5,  77.5])
        proj.longitude_of_central_meridian = -25.0
        proj.latitude_of_projection_origin = 77.5
        proj.earth_radius = 6371000.0
        proj.proj4 = '+proj=lcc +lat_0=77.5 +lon_0=-25 +lat_1=77.5 +lat_2=77.5 +no_defs +R=6.371e+06'

        #crs = ds.createVariable('spatial_ref', 'i4')
        #crs.spatial_ref='GEOGCS[PROJCS["unnamed",GEOGCS["unnamed ellipse",DATUM["unknown",SPHEROID["unnamed",6371000,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Lambert_Conformal_Conic_2SP"],PARAMETER["standard_parallel_1",77.5],PARAMETER["standard_parallel_2",77.5],PARAMETER["latitude_of_origin",77.5],PARAMETER["central_meridian",-25],PARAMETER["false_easting",0],PARAMETER["false_northing",0]]]'

        ds.close()
        os.close(fd) # Just in case - see https://www.logilab.org/blogentry/17873

        fd, self.filename_exported = tempfile.mkstemp(suffix='.nc')
        os.close(fd)

    def tearDown(self):
        super(TestExporter__export2thredds, self).tearDown()
        os.unlink(self.tmp_ncfile)
        os.unlink(self.filename_exported)

    def test_example1(self):
        n = Nansat(self.tmp_ncfile)
        n.export2thredds(self.filename_exported, mask_name='mask', no_mask_value=1)
        self.assertTrue(os.path.exists(self.filename_exported))
        ds = Dataset(self.filename_exported)
        self.assertEqual(ds.variables['x_wind_10m'][0,0,0].data, 9)
        self.assertTrue(np.isnan(ds.variables['x_wind_10m'][0,-1,0].data))

    def test_example2(self):
        n = Nansat(self.tmp_ncfile)
        res = n.export2thredds(self.filename_exported, {'x_wind_10m': {'description': 'example'}})
        self.assertEqual(res, None)

    def test_example3(self):
        n = Nansat(self.tmp_ncfile)
        res = n.export2thredds(self.filename_exported, {
            'x_wind_10m': {'type': '>i2', 'scale': 0.1, 'offset': 0},
            'y_wind_10m': {'type': '>i2', 'scale': 0.1, 'offset': 0}
        })
        # TODO: test that the type, scale and offset are actually modified according to the input
        self.assertEqual(res, None)

if __name__ == "__main__":
    unittest.main()



# ------------------------------------------------------------------------------
# Name:         test_pointbrowser.py
# Purpose:      Test the PointBrowser class
#
# Author:       Aleksander Vines, Anton Korosov
#
# Created:      2015-10-22
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
# ------------------------------------------------------------------------------
import os
import unittest
from mock import patch, PropertyMock, Mock, MagicMock, DEFAULT

import numpy as np
from nansat.pointbrowser import PointBrowser

try:
    import matplotlib
    import matplotlib.pyplot as plt
    plt.switch_backend('Agg')
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True


class PointBrowserTest(unittest.TestCase):
    @unittest.skipUnless(MATPLOTLIB_IS_INSTALLED, 'Matplotlib is required')
    def setUp(self):
        self.data = np.zeros((4, 4))
        self.event = MagicMock()

    def test_init(self):
        """ Create Pointbrowser """
        pb = PointBrowser(self.data, force_interactive=False)
        self.assertIsInstance(pb.fig, plt.Figure)
        self.assertTrue(np.alltrue(pb.data == self.data))
        self.assertTrue(np.alltrue(pb.ax.get_images()[0].get_array() == self.data))
        self.assertEqual(pb.fmt, 'x-k')
        self.assertEqual(pb.points, [])
        self.assertEqual(pb.coordinates, [[]])

    def test_onclick(self):
        """ Mimic click """
        self.event = MagicMock()
        self.event.xdata = 10
        self.event.ydata = 10
        self.event.key = None
        pb = PointBrowser(self.data, force_interactive=False)

        pb.onclick(self.event)
        self.assertIsInstance(pb.points[0][0], matplotlib.lines.Line2D)
        self.assertEqual(pb.coordinates, [[(self.event.xdata, self.event.ydata)]])

    def test_onclick_none(self):
        """ Mimic click outside figure  """
        self.event.xdata = None
        self.event.ydata = None
        self.event.key = None
        pb = PointBrowser(self.data, force_interactive=False)

        pb.onclick(self.event)
        self.assertEqual(pb.points, [])
        self.assertEqual(pb.coordinates, [[]])

    def test_onclick_key_z(self):
        """ Mimic click with 'z' pressed """
        self.event.xdata = 10
        self.event.ydata = 10
        self.event.key = 'z'
        pb = PointBrowser(self.data, force_interactive=False)

        pb.onclick(self.event)
        self.assertEqual(pb.points, [])
        self.assertEqual(pb.coordinates, [[]])

    def test_onclick_key(self):
        """ Mimic click with 'anykey' pressed """
        self.event = MagicMock()
        self.event.xdata = 10
        self.event.ydata = 10
        self.event.key = 'newkey'
        pb = PointBrowser(self.data, force_interactive=False)

        pb.onclick(self.event)
        self.assertIsInstance(pb.points[0][0], matplotlib.lines.Line2D)
        self.assertEqual(pb.coordinates, [[],[(self.event.xdata, self.event.ydata)]])

    def test_convert_coordinates(self):
        """ Mimic click with 'anykey' pressed """
        pb = PointBrowser(self.data, force_interactive=False)
        pb.coordinates = [[[1,2,3],[4,5,6]]]
        new_coordinates = pb._convert_coordinates()
        self.assertTrue(np.all(new_coordinates[0] == np.array([[1,2,3], [4,5,6]]).T))

    @patch('nansat.pointbrowser.plt')
    def test_get_points(self, plt_mock):
        plt_mock.show.return_value = None
        pb = PointBrowser(self.data, force_interactive=False)
        points = pb.get_points()
        self.assertTrue(pb.fig.canvas.mpl_connect.called)
        self.assertTrue(plt_mock.show.called)
        self.assertEqual(points, [])


if __name__ == "__main__":
    unittest.main()



# Name:        mapper_metno_hfr
# Purpose:     Mapper for CODAR SeaSonde High-Frequency radar data provided by MET Norway
#              For data see: https://thredds.met.no/thredds/catalog/remotesensinghfradar/catalog.html
# Authors:     Artem Moiseev
# Licence:     This file is part of NANSAT. You can redistribute it or modify
#              under the terms of GNU General Public License, v.3
#              http://www.gnu.org/licenses/gpl-3.0.html

import os
from datetime import datetime, timedelta
import json
from netCDF4 import Dataset, num2date
import numpy as np
from osgeo import gdal, osr
import pythesint as pti
from scipy.interpolate import griddata
from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    
    BAND_NAMES = ['direction', 'ersc', 'ertc', 'espc', 'etmp', 'maxv',
                  'minv', 'sprc', 'u', 'v', 'velo', 'vflg', 'xdst', 'ydst']
    SUPPORTED_LOCATIONS = ['RDLm_TORU', 'RDLm_FRUH', 'RDLm_BERL']

    def __init__(self, filename, gdal_dataset, gdal_metadata, GCP_COUNT=10, timestamp=None, **kwargs):
        filename_name = os.path.split(filename)[-1].split('.')[0]
        # Check if correct mapper
        correct_mapper = False 
        for location in self.SUPPORTED_LOCATIONS:
            # If it matches with one of locateions break the loop and flag True
            if filename_name.startswith(location):
                correct_mapper = True
                break
        if not correct_mapper:
            raise WrongMapperError

        # Import NetCDF4 dataset
        nc_dataset = Dataset(filename)
        # Define projection (depending on the HFR)
        if nc_dataset.getncattr('site') == 'TORU':
            proj4 = '+proj=utm +zone=32 +ellps=WGS84 +datum=WGS84 +units=m +no_defs'
            GRID_PX_SIZE = 1500 # Final raster px size in meters
        elif nc_dataset.getncattr('site') == 'FRUH':
            proj4 = '+proj=utm +zone=34 +ellps=WGS84 +datum=WGS84 +units=m +no_defs'
            GRID_PX_SIZE = 5000 # Final raster px size in meters
        elif nc_dataset.getncattr('site') == 'BERL':
            proj4 = '+proj=utm +zone=35 +ellps=WGS84 +datum=WGS84 +units=m +no_defs'
            GRID_PX_SIZE = 5000 # Final raster px size in meters
        else:
            raise WrongMapperError

        srs = osr.SpatialReference()
        srs.ImportFromProj4(proj4)
        projection = srs.ExportToWkt()
        # Get x grid and y grid
        x_grd, y_grd = self.create_linear_grid(nc_dataset['x'][:], nc_dataset['y'][:], GRID_PX_SIZE)
        raster_x_size, raster_y_size = x_grd.shape
        # Define geotransform
        geotransform = (x_grd.min(), GRID_PX_SIZE, 0.0, y_grd.max(), 0.0, GRID_PX_SIZE * -1)
        # Define x and y size
        self._init_from_dataset_params(raster_x_size, raster_y_size, geotransform, projection)
        # If required timestamp was not specified then extract date from filename and use first time
        if timestamp is None:
            timestamp = self.date_from_filename(filename)
        # Comvert time info from the dataset to the datetime
        timestamps = num2date(nc_dataset['time'][:].data, nc_dataset['time'].units)
        # find band id for the required timestamp
        # Note add 1 because in gdal counting starts from 1 not from 0
        src_timestamp_id = np.where(timestamps == timestamp)[0][0] + 1
        # Iterate through all subdatasets and bands to the dataset
        for subdataset in gdal_dataset.GetSubDatasets():
            # Get name of subdataset
            subdataset_name = subdataset[0].split(':')[2]
            # Check if the subdataset in the accepted 3D vars list
            if subdataset_name not in self.BAND_NAMES:
                continue
            gdal_subdataset = gdal.Open(subdataset[0])
            # need to be float for the nan replasement
            band_data = gdal_subdataset.GetRasterBand(int(src_timestamp_id)).ReadAsArray().astype('float')
            # remove fill value (replace with nan)
            fill_value = int(gdal_subdataset.GetMetadata_Dict()['#'.join([subdataset_name, '_FillValue'])])
            band_data[band_data == fill_value] = np.nan
            # Interpolate data on the regular grid
            band_grid_data = self.band2grid((nc_dataset['x'][:], nc_dataset['y'][:]), 
                                             band_data, (x_grd, y_grd))
            # Create VRT ffor the regridded data
            band_vrt = VRT.from_array(band_grid_data)
            # Add VRT to the list of all dataset vrts
            self.band_vrts[subdataset_name + 'VRT'] = band_vrt
            # Add band to the dataset
            src = {'SourceFilename': self.band_vrts[subdataset_name + 'VRT'].filename, 
                   'SourceBand': 1}
            # Add band specific metadata
            dst = {'name': subdataset_name}
            for key in gdal_subdataset.GetMetadata_Dict().keys():
                if key.startswith(subdataset_name):
                    clean_metadata_name = key.split('#')[1]
                    dst[clean_metadata_name] = gdal_subdataset.GetMetadata_Dict()[key]
            # Create band
            self.create_band(src, dst)
            self.dataset.FlushCache()

        # Set GCMD metadata
        self.dataset.SetMetadataItem('instrument', json.dumps(pti.get_gcmd_instrument('SCR-HF')))
        self.dataset.SetMetadataItem('platform', json.dumps(pti.get_gcmd_platform('CODAR SeaSonde')))
        self.dataset.SetMetadataItem('Data Center', json.dumps(pti.get_gcmd_provider('NO/MET')))
        self.dataset.SetMetadataItem('Entry Title', 'Near-Real Time Surface Ocean Radial Velocity')
        self.dataset.SetMetadataItem('gcmd_location',json.dumps(pti.get_gcmd_location('NORTH SEA')))
        # Set time coverage metadata
        self.dataset.SetMetadataItem('time_coverage_start', timestamp.isoformat())
        self.dataset.SetMetadataItem('time_coverage_end',
                                     (timestamp + timedelta(minutes=59, seconds=59)).isoformat())
        # Set NetCDF dataset metadata
        for key, value in gdal_dataset.GetMetadata_Dict().items():
            self.dataset.SetMetadataItem(key.split('#')[1], value)

    def date_from_filename(self, src):
        filename = os.path.splitext(os.path.basename(src))[0]
        year, month, day = filename.split('_')[-3:]
        src_timestamp = datetime(int(year), int(month), int(day), 0, 0, 0)
        return src_timestamp

    def create_linear_grid(self, x, y, px_size):
        x_grd, y_grd = np.meshgrid(np.arange(x.min(), x.max(), px_size),
                                   np.arange(y.max(), y.min(), px_size * -1))
        return x_grd, y_grd

    def band2grid(self, src_grd, var, dst_grd):
        # Points [(x, y), ... , ] from original file
        points = list(zip(src_grd[0].flatten(), src_grd[1].flatten()))
        repr_var = griddata(points, var.flatten(), (dst_grd[0], dst_grd[1]), method='nearest')
        return repr_var



# Name:    tools.py
# Purpose: Collection of functions which use core Nansat classes for
#          scientific data analysis and visualization
# Authors:      Artem Moiseev, Anton Korosov
# Created:      17.01.2020
# Copyright:    (c) NERSC 2020
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

import os
import warnings
import functools

try:
    import matplotlib.pyplot as plt
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True

try:
    import cartopy.crs as ccrs
except ImportError:
    CARTOPY_IS_INSTALLED = False
else:
    CARTOPY_IS_INSTALLED = True

from nansat.nansat import Nansat
from nansat import utils

def distance2coast(dst_domain, distance_src=None):
    """ Estimate distance to the nearest coast (in km) for each pixcel in the
    domain of interest. The method utilizes NASA's OBPG group Distance to the Nearest Coast
    product: https://oceancolor.gsfc.nasa.gov/docs/distfromcoast/. The product is stored in GeoTiff
    format with pixcelsize of 0.01x0.01 degree.

    Parameters
    -----------
    dst_domain : Domain
        destination domain
    distance_src : str
        path to the NASA Distance to the Nearest coast GeoTIFF product

    Returns
    --------
    distance : Nansat object with distance to the coast mask in current projection

    See Also
    ---------
    `<https://oceancolor.gsfc.nasa.gov/docs/distfromcoast/>`_
    `<http://nansat.readthedocs.io/en/latest/source/features.html#differentiating-between-land-and-water>`

    """
    # Get path to the auxilary dataset predefined in enviromental variable
    if distance_src is None:
        distance_src = os.getenv('DIST2COAST')
    # If path to the distance data source was not specified or directly provided raise an error
    if distance_src is None or not os.path.exists(distance_src):
        raise IOError('Distance to the nearest coast product does not exist - see Nansat '
                      'documentation to get it (the path is % s)' % distance_src)
    distance = Nansat(distance_src)
    # Reproject the source file on the domain of interest
    distance.reproject(dst_domain, addmask=False)
    return distance

def get_domain_map(domain,
                   crs=None,
                   lon_margin=10.,
                   lat_margin=10.,
                   lw=1,
                   linestyle='b.-',
                   fill_color='coral',
                   fill_alpha=0.5,
                   draw_gridlines=True,
                   draw_labels=True,
                   grid_lw=2,
                   grid_color='gray',
                   grid_alpha=0.5,
                   grid_linestyle='--',
                   **kwargs):
    """
    Create a pyplot figure axis with Domain map, Cartopy projection, coastlines

    Parameters
    -----------
    domain : Domain
        the desired Domain to plot
    crs : cartopy CRS or None
        projection of the map, cartopy.crs.PlateCaree by default
    lon_margin : float
        10, horisontal border around patch (degrees of longitude)
    lat_margin : float
        10, vertical border around patch (degrees of latitude)
    linestyle : str
        domain line style
    lw : float
        domain line width
    fill_color : str
        domain fill color
    fill_alpha : float
        domain fill transparency
    draw_gridlines : bool
        Add gridlines to the plot?
    draw_labels : bool
        Add labels to the plot?
    grid_lw : float
        gridlines line width
    grid_color : str,
        gridlines color
    grid_alpha : float
        gridlines transparency
    grid_linestyle : str
        gridlines style

    Returns
    -------
    ax : pyplot/cartopy axes
        Axes with the map and domain patch

    """
    if not CARTOPY_IS_INSTALLED:
        raise ImportError(' Cartopy is not installed. Cannot use get_domain_map '
                          ' Enable by: conda install -c conda-forge cartopy ')

    if crs is None:
        crs = ccrs.PlateCarree()

    blon, blat = domain.get_border()

    ax = plt.subplot(111, projection=crs)
    ax.stock_img()
    ax.coastlines()
    ax.plot(blon, blat, linestyle,
                        lw=lw,
                        transform=ccrs.PlateCarree())
    ax.fill(blon, blat, color=fill_color,
                        alpha=fill_alpha,
                        transform=crs)
    plt.xlim([blon.min()-lon_margin, blon.max()+lon_margin])
    plt.ylim([blat.min()-lat_margin, blat.max()+lat_margin])
    if draw_gridlines:
        gl = ax.gridlines(draw_labels=draw_labels,
                          linewidth=grid_lw,
                          color=grid_color,
                          alpha=grid_alpha,
                          linestyle=grid_linestyle,
                          crs=crs)
    return ax


def show_domain_map(domain, **kwargs):
    """ Show Domain map interactively

    Parameters
    ----------
    domain : Domain
        the Domain to show
    **kwargs : dict
        parameters for nansat.tools.get_domain_map

    """
    ax = get_domain_map(domain, **kwargs)
    plt.show()

def save_domain_map(domain,
                    filename,
                    figsize=(5,5),
                    dpi=150,
                    bbox_inches='tight',
                    pad_inches=0,
                    **kwargs):
    """ Save Domain to PNG file

    Parameters
    ----------
    domain : Domain
        the Domain to show
    filename : str
        destination filename
    figsize : (int, int)
        size of the figure in inches
    dpi : PNG resolution
    bbox_inches : str or int
        see pyplot.savefig
    pad_inches : int
        see pyplot.savefig
    **kwargs : dict
        parameters for nansat.tools.get_domain_map

    """
    plt.figure(figsize=figsize)
    ax = get_domain_map(domain, **kwargs)
    plt.savefig(filename, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close()

def deprecated(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn('The function <{}> was moved to the nansat.utils module and soon will not '
                      'be available from the nansat.tools'.format(func.__name__),
                      stacklevel=2)
        return func(*args, **kwargs)
    return new_func


@deprecated
def remove_keys(dict, keys):
    return utils.remove_keys(dict, keys)


@deprecated
def register_colormaps():
    utils.register_colormaps()


@deprecated
def initial_bearing(lon1, lat1, lon2, lat2):
    return utils.initial_bearing(lon1, lat1, lon2, lat2)


@deprecated
def haversine(lon1, lat1, lon2, lat2):
    return utils.haversine(lon1, lat1, lon2, lat2)


@deprecated
def add_logger(logName='', logLevel=None):
    return utils.add_logger(logName=logName, logLevel=logLevel)


@deprecated
def get_random_color(c0=None, minDist=100, low=0, high=255):
    return utils.get_random_color(c0=c0, minDist=minDist, low=low, high=high)


@deprecated
def parse_time(time_string):
    return utils.parse_time(time_string)



# -------------------------------------------------------------------------------
# Name:         mapper_topography.py
# Purpose:      Mapping for the global 30 arc-second elevation
#
# Author:       Morten Wergeland Hansen
# Modified:	Morten Wergeland Hansen
#
# Created:	04.06.2015
# Last modified:08.06.2015 10:27
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
# -------------------------------------------------------------------------------
from __future__ import division, unicode_literals, absolute_import, print_function
import re
import os.path

from nansat.vrt import VRT

from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    """
    Mapping for the GTOPO30 (`<https://lta.cr.usgs.gov/GTOPO30>`_) and the GMTED2010
    (`<https://lta.cr.usgs.gov/GMTED2010>`_) global elevation models.

    Parameters
    ----------
    filename : string
        The vrt filename, e.g., ``gtopo30.vrt``
    gdal_dataset : osgeo.gdal.Dataset
        The GDAL dataset returned by ``gdal.Open(filename)``

    Example
    -------
    You can create your own ``gtopo30.vrt`` file with gdal, e.g.:

    .. code-block:: bash

        gdalbuildvrt <dem>.vrt [E,W]*.DEM

    Note
    ----
    Either the name of a GTOPO30 DEM file or GMTED2010 tif file, or ``<path>/<dem>.vrt``. The latter
    is an aggregation of the DEM-files available from the given DEM. The GTOPO30 vrt does not
    contain the Antarctic, because this is in polarstereographic projection.

    Remember to update this mapper by adding allowed filenames to the list of accepted filenames
    (accepted_names) if you create or apply new DEM datasets.

    """

    def __init__(self, filename, gdal_dataset, *args, **kwargs):

        bn = os.path.basename(filename)
        accepted_names = [
                'gmted2010_30.vrt',
                'gtopo30.vrt',
                '.*.DEM',
                '.*_gmted_mea.*.tif',
            ]

        correct_mapper = False
        for accepted_name in accepted_names:
            m = re.search(accepted_name, filename)
            if m and m.group(0):
                correct_mapper = True
                break

        #if not bn=='gtopo30.vrt' and not os.path.splitext(bn)[1]=='.DEM':
        if not correct_mapper:
            raise WrongMapperError

        metaDict = [{'src': {'SourceFilename': filename, 'SourceBand':  1},
                     'dst': {'wkv': 'height_above_reference_ellipsoid'}}]

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdal_dataset)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)



from __future__ import absolute_import, unicode_literals, division
import unittest
import sys
import datetime
import json

import pythesint as pti

from nansat import Nansat
from nansat.nansat import _import_mappers
from nansat_integration_tests.mapper_test_archive import DataForTestingMappers, DataForTestingOnlineMappers


class TestDataForTestingMappers(unittest.TestCase):
    def test_create_test_data(self):
        ''' should create TestData instance '''
        t = DataForTestingMappers()
        self.assertTrue(hasattr(t, 'mapperData'))


# https://nose.readthedocs.org/en/latest/writing_tests.html#test-generators
# The x-flag results in the test stopping at first failure or error - use it
# for easier debugging:
# nosetests -v -x integration_tests.test_mappers:TestAllMappers.test_mappers_basic
class TestAllMappers(object):

    @classmethod
    def setup_class(cls):
        ''' Download testing data '''
        cls.testData = DataForTestingMappers()

    def test_mappers_basic(self):
        ''' Basic mapper test '''
        for kwargs in self.testData.mapperData:
            fileName = kwargs.pop('fileName')
            mapperName = kwargs.pop('mapperName')
            # Test call to Nansat, mapper not specified
            sys.stderr.write('\n -> '+fileName+'\n')
            yield self.open_with_nansat, fileName, None, kwargs
            # Test call to Nansat, mapper specified
            sys.stderr.write('\n'+mapperName+' -> '+fileName+'\n')
            yield self.open_with_nansat, fileName, mapperName, kwargs

    def open_with_nansat(self, filePath, mapper=None, kwargs=None):
        ''' Ensures that you can open the filePath as a Nansat object '''
        if kwargs is None:
            kwargs = {}

        try:
            if mapper:
                n = Nansat(filePath, mapperName=mapper, **kwargs)
            else:
                n = Nansat(filePath, **kwargs)
        except Exception as e:
            raise Exception('%s: %s'%(filePath, e.message))
        assert type(n) == Nansat

    def test_mappers_advanced(self):
        ''' Run tests to check DIF/GCMD metadata content in all mappers'''
        for dd in self.testData.mapperData:
            sys.stderr.write('\nMapper '+dd['mapperName']+' -> '+dd['fileName']+'\n')
            n = Nansat(dd['fileName'], mapperName=dd['mapperName'])
            yield self.is_correct_mapper, n, dd['mapperName']
            yield self.has_metadata_time_coverage_start, n
            yield self.has_metadata_time_coverage_end, n
            yield self.has_metadata_platform, n
            yield self.has_metadata_instrument, n

            # Test that SAR objects have sigma0 intensity bands in addition
            # to complex bands
            if n.has_band(
                'surface_backwards_scattering_coefficient_of_radar_wave'
                    ):
                yield self.exist_intensity_band, n

    def has_metadata_time_coverage_start(self, n):
        ''' Has start time '''
        assert type(n.time_coverage_start) == datetime.datetime

    def has_metadata_time_coverage_end(self, n):
        assert type(n.time_coverage_end) == datetime.datetime

    def has_metadata_platform(self, n):
        meta1 = json.loads(n.get_metadata('platform'))
        meta1ShortName = meta1['Short_Name']
        meta2 = pti.get_gcmd_platform(meta1ShortName)

        assert type(meta1) == dict
        assert meta1 == meta2

    def has_metadata_instrument(self, n):
        meta1 = json.loads(n.get_metadata('instrument'))
        meta1ShortName = meta1['Short_Name']
        meta2 = pti.get_gcmd_instrument(meta1ShortName)

        assert type(meta1) == dict
        assert meta1 == meta2

    def is_correct_mapper(self, n, mapper):
        assert n.mapper == mapper

    def exist_intensity_band(self, n):
        ''' Test if intensity bands exist for complex data '''
        allBandNames = []
        complexBandNames = []
        for iBand in range(n.vrt.dataset.RasterCount):
            iBandName = n.get_metadata(band=iBand + 1)['name']
            allBandNames.append(iBandName)
            if '_complex' in iBandName:
                complexBandNames.append(iBandName)

        for iComplexName in complexBandNames:
            assert iComplexName.replace('_complex', '') in allBandNames

class TestOnlineMappers(TestAllMappers):
    @classmethod
    def setup_class(cls):
        ''' Download testing data '''
        cls.testData = DataForTestingOnlineMappers()


if __name__ == '__main__':
    # nansatMappers = _import_mappers()
    # for mapper in nansatMappers:
    #     test_name = 'test_%s'%mapper
    unittest.main()




# Name:         mapper_case2reg.py
# Purpose:      Mapper for the BEAM/Visat output of Case2Regional algorithm
# Authors:      Asuka Yamakava, Anton Korosov, Morten Wergeland Hansen
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os.path
import numpy as np

from nansat.vrt import VRT
from nansat.node import Node
from nansat.nsr import NSR
from nansat.exceptions import WrongMapperError

import pythesint as pti

from nansat.mappers import mapper_generic as mg

class Mapper(mg.Mapper):
    '''Mapping for the BEAM/Visat output of Case2Regional algorithm'''
    def __init__(self, filename, gdalDataset, gdalMetadata,
                 wavelengths=[None, 413, 443, 490, 510, 560, 620, 665,
                              681, 709, 753, None, 778, 864], **kwargs):

        fPathName, fExt = os.path.splitext(filename)
        fPath, fName = os.path.split(fPathName)
        if fExt != '.nc' or 'MER_' not in fName or 'N1_C2IOP' not in fName:
            raise WrongMapperError

        # get all metadata using the GENERIC Mapper
        mg.Mapper.__init__(self, filename, gdalDataset, gdalMetadata)

        #add metadata for Rrs bands
        rrsDict = pti.get_wkv_variable('surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air')

        for bi in range(1, 1+self.dataset.RasterCount):
            b = self.dataset.GetRasterBand(bi)
            bMetadata = b.GetMetadata()
            rawName = bMetadata.get('name', '')
            if 'reflec_' in rawName:
                refNumber = int(rawName.split('_')[1])
                wavelength = wavelengths[refNumber]
                b.SetMetadataItem('name', 'Rrs_' + str(wavelength))
                b.SetMetadataItem('wavelength', str(wavelength))
                for rrsKey in rrsDict:
                    b.SetMetadata(rrsKey, rrsDict[rrsKey])

                src = [{
                    'SourceFilename': b.GetMetadataItem('SourceFilename'),
                    'SourceBand':  b.GetMetadataItem('SourceBand'),
                    'DataType': 6}]
                dst = {
                    'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_water',
                    'suffix': str(wavelength),
                    'wavelength': str(wavelength),
                    'PixelFunctionType': 'NormReflectanceToRemSensReflectance'}
                self.create_band(src, dst)

        self.dataset.FlushCache()



# Name:        obpg
# Purpose:     Base class for mapping for L2 data from the OBPG web-site
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from nansat.vrt import VRT


class OBPGL2BaseClass(VRT):
    ''' Base Class for Mappers for SeaWIFS/MODIS/MERIS/VIIRS L2 data from OBPG
    '''

    titles = ['HMODISA Level-2 Data',
              'MODISA Level-2 Data',
              'HMODIST Level-2 Data',
              'MERIS Level-2 Data',
              'GOCI Level-2 Data',
              'VIIRSN Level-2 Data',
              'SeaWiFS Level-2 Data']



# Name:  nansat.py
# Purpose: Container of Nansat class
# Authors:      Anton Korosov, Knut-Frode Dagestad, Morten W. Hansen, Artem Moiseev,
#               Asuka Yamakawa, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2018
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import absolute_import, print_function
import os
import glob
import sys
import tempfile
import datetime
import pkgutil
import warnings
from xml.sax import saxutils

import numpy as np
from numpy import nanmedian
from numpy.lib.recfunctions import append_fields
from netCDF4 import Dataset

try:
    from matplotlib import cm
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True

from nansat.domain import Domain
from nansat.exporter import Exporter
from nansat.figure import Figure
from nansat.vrt import VRT
from nansat.utils import add_logger, gdal, parse_time
from nansat.node import Node
from nansat.pointbrowser import PointBrowser

from nansat.exceptions import NansatGDALError, WrongMapperError, NansatReadError

import collections
if hasattr(collections, 'OrderedDict'):
    from collections import OrderedDict
else:
    from ordereddict import OrderedDict

# container for all mappers
nansatMappers = None


class Nansat(Domain, Exporter):
    """Container for geospatial data. Performs all high-level operations.

    n = Nansat(filename) opens the file with satellite or model data for
    reading, adds scientific metadata to bands, and prepares the data for
    further handling.

    Parameters
    -----------
    filename : str
        uri of the input file or OpeNDAP datastream
    mapper : str
        name of the mapper from nansat/mappers dir. E.g.
        'sentinel1_l1', 'asar', 'hirlam', 'meris_l1', 'meris_l2', etc.
    log_level : int
        Level of logging. See: http://docs.python.org/howto/logging.html
    kwargs : additional arguments for mappers

    Examples
    --------
        >>> n1 = Nansat(filename)
        >>> n2 = Nansat(sentinel1_filename, mapper='sentinel1_l1')
        >>> array1 = n1[1]
        >>> array2 = n2['sigma0_HV']

    Notes
    -----
    The instance of Nansat class (the object <n>) contains information
    about geographical reference of the data (e.g raster size, pixel
    resolution, type of projection, etc) and about bands with values of
    geophysical variables (e.g. water leaving radiance, normalized radar
    cross section, chlrophyll concentraion, etc). The object <n> has methods
    for high-level operations with data. E.g.:
    * reading data from file (Nansat.__getitem__);
    * visualization (Nansat.write_figure);
    * changing geographical reference (Nansat.reproject);
    * exporting (Nansat.export)
    * and much more...

    Nansat inherits from Domain (container of geo-reference information)
    Nansat uses instance of VRT (wraper around GDAL VRT-files)
    Nansat uses instance of Figure (collection of methods for visualization)

    """

    FILL_VALUE = 9.96921e+36
    ALT_FILL_VALUE = -10000.

    # instance attributes
    logger = None
    filename = None
    name = None
    path = None
    vrt = None
    mapper = None

    @classmethod
    def from_domain(cls, domain, array=None, parameters=None, log_level=30):
        """Create Nansat object from input Domain [and array with data]

        Parameters
        ----------
        domain : Domain
            Defines spatial reference system and geographical extent.
        array : numpy NDarray
            Data for the first band. Shape must correspond to shape of <domain>
        parameters : dict
            Metadata for the first band. May contain 'name', 'wkv' and other keys.
        log_level : int
            Level of logging.

        """
        n = cls.__new__(cls)
        n._init_from_domain(domain, array, parameters, log_level)
        return n

    def __init__(self, filename='', mapper='', log_level=30, **kwargs):
        """Create Nansat object

        Notes
        -----
        self.mapper : str
            name of the used mapper
        self.filename : file name
            set file name given by the argument
        self.vrt : VRT
            Wrapper around VRT file and GDAL dataset with satellite raster data
        self.logger : logging.Logger
            logger for output debugging info
        self.name : str
            name of object (for writing KML)
        self.path : str
            path to input file

        """
        if filename == '':
            raise ValueError('Nansat is called without valid parameters! Use: Nansat(filename)')

        self._init_empty(filename, log_level)
        # Create VRT object with mapping of variables
        self.vrt = self._get_mapper(mapper, **kwargs)

    def __getitem__(self, band_id):
        """Returns the band as a NumPy array, by overloading []

        Parameters
        -----------
        band_id : int or str
            If int, array from band with number <band_id> is returned
            If string, array from band with metadata 'name' equal to
            <band_id> is returned

        Returns
        --------
        a : NumPy array


        """
        # get band
        band = self.get_GDALRasterBand(band_id)
        # get expression from metadata
        expression = band.GetMetadata().get('expression', '')
        # get data
        band_data = band.ReadAsArray()
        if band_data is None:
            raise NansatGDALError('Cannot read array from band %s' % str(band_data))

        # execute expression if any
        if expression != '':
            band_data = eval(expression)

        all_float_flag = band_data.dtype.char in np.typecodes['AllFloat']
        # Set invalid and missing data to np.nan (for floats only)
        if '_FillValue' in band.GetMetadata() and all_float_flag:
            band_data = self._fill_with_nan(band, band_data)

        # replace infs with np.NAN
        if np.size(np.where(np.isinf(band_data))) > 0:
            band_data[np.isinf(band_data)] = np.nan

        # erase out-of-swath pixels with np.Nan (if not integer)
        if self.has_band('swathmask') and all_float_flag:
            swathmask = self.get_GDALRasterBand('swathmask').ReadAsArray()
            band_data[swathmask == 0] = np.nan

        return band_data

    def __repr__(self):
        """Creates string with basic info about the Nansat object"""
        out_str = '{separator}{filename}{separator}Mapper: {mapper}{bands}{separator}{domain}'
        return out_str.format(separator=self.OUTPUT_SEPARATOR, filename=self.filename,
                              bands=self.list_bands(False), mapper=self.mapper,
                              domain=Domain.__repr__(self))

    def _init_empty(self, filename, log_level):
        """Init empty Nansat object

        Parameters
        ----------
        filename : str
            Name of input file.
        log_level : int
            Level of logging verbosity.

        Notes
        --------
            self.logger
                adds logging.Logger
            self.filename
                adds full path to input file
            self.name
                adds name of file
            self.path
                adds path to file

        """
        # create logger
        self.logger = add_logger('Nansat', log_level)
        # set input file name
        self.filename = filename
        # name, for compatibility with some Domain methods
        self.name = os.path.basename(filename)
        self.path = os.path.dirname(filename)

    def _init_from_domain(self, domain, array=None, parameters=None, log_level=30):
        """Init Nansat object from input Domain and optionally array with band values

        Parameters
        ----------
        domain : Domain
            Defines spatial reference system and geographical extent.
        array : numpy NDarray
            Data for the first band. Shape must correspond to shape of <domain>
        parameters : dict
            Metadata for the first band. May contain 'name', 'wkv' and other keys.
        log_level : int
            Level of logging.

        """
        self._init_empty('', log_level)
        self.vrt = VRT.from_gdal_dataset(domain.vrt.dataset, geolocation=domain.vrt.geolocation)
        self.mapper = ''
        if array is not None:
            self.add_band(array=array, parameters=parameters)

    def _fill_with_nan(self, band, band_data):
        """Fill input array with fill value taen from input band metadata"""
        fill_value = float(band.GetMetadata()['_FillValue'])
        band_data[band_data == fill_value] = np.nan
        # quick hack to avoid problem with wrong _FillValue - see issue
        # #123
        if fill_value == self.FILL_VALUE:
            band_data[band_data == self.ALT_FILL_VALUE] = np.nan

        return band_data


    def add_band(self, array, parameters=None, nomem=False):
        """Add band from numpy array with metadata.

        Create VRT object which contains VRT and RAW binary file and append it
        to self.vrt.band_vrts

        Parameters
        -----------
        array : ndarray
            new band data. Shape should be equal to shape
        parameters : dict
            band metadata: wkv, name, etc. (or for several bands)
        nomem : bool
            saves the vrt to a tempfile on disk?

        Notes
        -----
        Creates VRT object with VRT-file and RAW-file. Adds band to the self.vrt.

        Examples
        --------
            >>> n.add_band(array, {'name': 'new_data'}) # add new band and metadata, keep in memory
            >>> n.add_band(array, nomem=True) # add new band, keep on disk

        """
        self.add_bands([array], [parameters], nomem)

    def add_bands(self, arrays, parameters=None, nomem=False):
        """Add bands from numpy arrays with metadata.

        Create VRT object which contains VRT and RAW binary file and append it
        to self.vrt.band_vrts

        Parameters
        -----------
        arrays : list of ndarrays
            new band data. Shape should be equal to shape
        parameters : list of dict
            band metadata: wkv, name, etc. (or for several bands)
        nomem : bool
            saves the vrt to a tempfile on disk?

        Notes
        -----
        Creates VRT object with VRT-file and RAW-file. Adds band to the self.vrt.

        Examples
        --------
            >>> n.add_bands([array1, array2]) # add new bands, keep in memory

        """
        # replace empty parameters with list of empty dictionaries
        if parameters is None:
            parameters = [{}] * len(arrays)

        self.vrt = self.vrt.get_super_vrt()

        # create VRTs from arrays and generate band_metadata
        band_metadata = []
        for array, parameter in zip(arrays, parameters):
            vrt = VRT.from_array(array, nomem=nomem)
            band_metadata.append({
                'src': {'SourceFilename': vrt.filename, 'SourceBand': 1},
                'dst': parameter
                })
            self.vrt.band_vrts[vrt.filename] = vrt

        band_name = self.vrt.create_bands(band_metadata)

    def bands(self):
        """Make a dictionary with all metadata from all bands

        Returns
        --------
        b : dictionary
            key = N, value = dict with all band metadata

        """
        band_metadata = {}
        for band_num in range(self.vrt.dataset.RasterCount):
            band_metadata[band_num + 1] = self.get_metadata(band_id=band_num + 1)

        return band_metadata

    def has_band(self, band):
        """Check if self has band with name <band>
        Parameters
        ----------
            band : str
                name or standard_name of the band to check

        Returns
        -------
            True/False if band exists or not

        """
        for band_num in self.bands():
            band_meta = self.bands()[band_num]
            if band_meta['name'] == band:
                return True
            elif 'standard_name' in band_meta and band_meta['standard_name'] == band:
                return True

    def _get_resize_shape(self, factor, width, height, dst_pixel_size):
        """Estimate new shape either from factor or destination width/height or pixel size"""
        src_shape = np.array(self.shape(), float)
        # estimate factor if either width or height is given and factor is not given
        if width is not None:
            factor = width / src_shape[1]
        if height is not None:
            factor = height / src_shape[0]

        # estimate factor if pixelsize is given
        if dst_pixel_size is not None:
            src_pixel_size = np.array(self.get_pixelsize_meters(), float)[::-1]
            factor = (src_pixel_size / float(dst_pixel_size)).mean()

        factor = float(factor)
        dst_shape = np.floor(src_shape * factor)
        self.logger.info('New shape: ({0}, {1})'.format(dst_shape[0], dst_shape[1]))
        return factor, dst_shape

    def resize(self, factor=None, width=None, height=None, pixelsize=None, resample_alg=-1):
        """Proportional resize of the dataset.

        The dataset is resized as (x_size*factor, y_size*factor)
        If desired width, height or pixelsize is specified,
        the scaling factor is calculated accordingly.
        If GCPs are given in a dataset, they are also rewritten.

        Parameters
        -----------
        factor : float, optional, default=1
            Scaling factor for width and height

            - > 1 means increasing domain size
            - < 1 means decreasing domain size

        width : int, optional
            Desired new width in pixels
        height : int, optional
            Desired new height in pixels
        pixelsize : float, optional
            Desired new pixelsize in meters (approximate).
            A factor is calculated from ratio of the
            current pixelsize to the desired pixelsize.
        resample_alg : int (GDALResampleAlg), optional

            - -1 : Average (default),
            - 0 : NearestNeighbour
            - 1 : Bilinear,
            - 2 : Cubic,
            - 3 : CubicSpline,
            - 4 : Lancoz

        Notes
        -----
        self.vrt.dataset : VRT dataset of VRT object
            raster size are modified to downscaled size.
            If GCPs are given in the dataset, they are also overwritten.

        """
        factor, dst_shape = self._get_resize_shape(factor, width, height, pixelsize)
        if resample_alg <= 0:
            self.vrt = self.vrt.get_subsampled_vrt(dst_shape[1], dst_shape[0], resample_alg)
        else:
            # update size and GeoTranform in XML of the warped VRT object
            self.vrt = self.vrt.get_resized_vrt(dst_shape[1], dst_shape[0], resample_alg)

        # resize gcps
        gcps = self.vrt.vrt.dataset.GetGCPs()
        if len(gcps) > 0:
            gcpPro = self.vrt.vrt.dataset.GetGCPProjection()
            for gcp in gcps:
                gcp.GCPPixel *= factor
                gcp.GCPLine *= factor
            self.vrt.dataset.SetGCPs(gcps, gcpPro)
            self.vrt._remove_geotransform()
        else:
            # change resultion in geotransform to keep spatial extent
            geoTransform = list(self.vrt.vrt.dataset.GetGeoTransform())
            geoTransform[1] = float(geoTransform[1])/factor
            geoTransform[5] = float(geoTransform[5])/factor
            geoTransform = list(map(float, geoTransform))
            self.vrt.dataset.SetGeoTransform(geoTransform)

        # set global metadata
        subMetaData = self.vrt.vrt.dataset.GetMetadata()
        subMetaData.pop('filename')
        self.set_metadata(subMetaData)

        return factor

    def get_GDALRasterBand(self, band_id=1):
        """Get a GDALRasterBand of a given Nansat object

        If str is given find corresponding band number
        If int is given check if band with this number exists.
        Get a GDALRasterBand from vrt.

        Parameters
        -----------
        band_id : int or str
            - if int - a band number of the band to fetch
            - if str band_id = {'name': band_id}

        Returns
        --------
        GDAL RasterBand

        Example
        -------
            >>> b = n.get_GDALRasterBand(1)
            >>> b = n.get_GDALRasterBand('sigma0')

        """
        # get band number
        bandNumber = self.get_band_number(band_id)
        # the GDAL RasterBand of the corresponding band is returned
        return self.vrt.dataset.GetRasterBand(bandNumber)

    def list_bands(self, do_print=True):
        """Show band information of the given Nansat object

        Show serial number, longName, name and all parameters
        for each band in the metadata of the given Nansat object.

        Parameters
        -----------
        do_print : boolean
            print on screen?

        Returns
        --------
        outString : String
            formatted string with bands info

        """
        # get dictionary of bands metadata
        bands = self.bands()
        outString = ''

        for b in bands:
            # print band number, name
            outString += 'Band : %d %s\n' % (b, bands[b].get('name', ''))
            # print band metadata
            for i in bands[b]:
                outString += '  %s: %s\n' % (i, bands[b][i])
        if do_print:
            # print to screeen
            print(outString)
        else:
            return outString

    def reproject(self, dst_domain=None, resample_alg=0,
                  block_size=None, tps=None, skip_gcps=1, addmask=True,
                  **kwargs):
        """
        Change projection of the object based on the given Domain

        Create superVRT from self.vrt with AutoCreateWarpedVRT() using
        projection from the dst_domain.
        Modify XML content of the warped vrt using the Domain parameters.
        Generate warpedVRT and replace self.vrt with warpedVRT.
        If current object spans from 0 to 360 and dst_domain is west of 0,
        the object is shifted by 180 westwards.

        Parameters
        -----------
        dst_domain : domain
            destination Domain where projection and resolution are set
        resample_alg : int (GDALResampleAlg)

            - 0 : NearestNeighbour
            - 1 : Bilinear
            - 2 : Cubic,
            - 3 : CubicSpline
            - 4 : Lancoz

        block_size : int
            size of blocks for resampling. Large value decrease speed
            but increase accuracy at the edge
        tps : bool
            Apply Thin Spline Transformation if source or destination has GCPs
            Usage of TPS can also be triggered by setting self.vrt.tps=True
            before calling to reproject.
            This options has priority over self.vrt.tps
        skip_gcps : int
            Using TPS can be very slow if the number of GCPs are large.
            If this parameter is given, only every [skip_gcp] GCP is used,
            improving calculation time at the cost of accuracy.
            If not given explicitly, 'skip_gcps' is fetched from the
            metadata of self, or from dst_domain (as set by mapper or user).
            [defaults to 1 if not specified, i.e. using all GCPs]
        addmask : bool
            If True, add band 'swathmask'. 1 - valid data, 0 no-data.
            This band is used to replace no-data values with np.nan

        Notes
        -----
        self.vrt : VRT object with dataset replaced to warpedVRT dataset
        Integer data is returnd by integer. Round off to decimal place.
        If you do not want to round off, convert the data types to
        GDT_Float32, GDT_Float64, or GDT_CFloat32.

        See Also
        ---------
        `<http://www.gdal.org/gdalwarp.html>`_

        """
        # This is time consuming and therefore not done...:
        #if not self.overlaps(dst_domain):
        #    raise ValueError('Source and destination domains do not overlap')

        # if self spans from 0 to 360 AND dst_domain is west of 0:
        #     shift self westwards by 180 degrees
        # check span
        srcCorners = self.get_corners()
        if round(min(srcCorners[0])) == 0 and round(max(srcCorners[0])) == 360:
            # check intersection of src and dst
            dstCorners = dst_domain.get_corners()
            if min(dstCorners[0]) < 0:
                # shift
                self.vrt = self.vrt.get_shifted_vrt(-180)

        # get projection of destination dataset
        dstSRS = dst_domain.vrt.dataset.GetProjection()

        # get destination GCPs
        dstGCPs = dst_domain.vrt.dataset.GetGCPs()
        if len(dstGCPs) > 0:
            # get projection of destination GCPs
            dstSRS = dst_domain.vrt.dataset.GetGCPProjection()

        x_size = dst_domain.vrt.dataset.RasterXSize
        y_size = dst_domain.vrt.dataset.RasterYSize

        geoTransform = dst_domain.vrt.dataset.GetGeoTransform()

        # set trigger for using TPS
        if tps is True:
            self.vrt.tps = True
        elif tps is False:
            self.vrt.tps = False

        # Reduce number of GCPs for faster reprojection
        # when using TPS (if requested)
        src_skip_gcps = self.vrt.dataset.GetMetadataItem('skip_gcps')
        dst_skip_gcps = dst_domain.vrt.dataset.GetMetadataItem('skip_gcps')
        kwargs['skip_gcps'] = skip_gcps  # default (use all GCPs)
        if dst_skip_gcps is not None:  # ...or use setting from dst
            kwargs['skip_gcps'] = int(dst_skip_gcps)
        if src_skip_gcps is not None:  # ...or use setting from src
            kwargs['skip_gcps'] = int(src_skip_gcps)

        # add band that masks valid values with 1 and nodata with 0
        # after reproject
        # TODD: REFACTOR: replace with VRT._add_swath_mask_band
        if addmask:
            self.vrt = self.vrt.get_super_vrt()
            src = [{
                'SourceFilename': self.vrt.vrt.filename,
                'SourceBand':  1,
                'DataType': gdal.GDT_Byte
            }]
            dst = {
                'dataType': gdal.GDT_Byte,
                'wkv': 'swath_binary_mask',
                'PixelFunctionType': 'OnesPixelFunc',
            }
            self.vrt.create_band(src=src, dst=dst)
            self.vrt.dataset.FlushCache()

        # create Warped VRT
        self.vrt = self.vrt.get_warped_vrt(dstSRS, x_size, y_size, geoTransform,
                                           resample_alg=resample_alg,
                                           dst_gcps=dstGCPs,
                                           block_size=block_size, **kwargs)

        # set global metadata from subVRT
        subMetaData = self.vrt.vrt.dataset.GetMetadata()
        subMetaData.pop('filename')
        self.set_metadata(subMetaData)

    def undo(self, steps=1):
        """Undo reproject, resize, add_band or crop of Nansat object

        Restore the self.vrt from self.vrt.vrt

        Parameters
        -----------
        steps : int
            How many steps back to undo

        Notes
        ------
        Modifies self.vrt

        """
        self.vrt = self.vrt.get_sub_vrt(steps)

    def watermask(self, mod44path=None, dst_domain=None, **kwargs):
        """
        Create numpy array with watermask (water=1, land=0)

        250 meters resolution watermask from MODIS 44W Product:
        http://www.glcf.umd.edu/data/watermask/

        Watermask is stored as tiles in TIF(LZW) format and a VRT file
        All files are stored in one directory.
        A tarball with compressed TIF and VRT files should be additionally
        downloaded from the Nansat documentation page:
        http://nansat.readthedocs.io/en/latest/source/features.html#differentiating-between-land-and-water

        The method :
            Gets the directory either from input parameter or from environment
            variable MOD44WPATH
            Open Nansat object from the VRT file
            Reprojects the watermask onto the current object using reproject()
            or reproject_on_jcps()
            Returns the reprojected Nansat object

        Parameters
        -----------
        mod44path : string
            path with MOD44W Products and a VRT file
        dst_domain : Domain
            destination domain other than self
        tps : Bool
            Use Thin Spline Transformation in reprojection of watermask?
            See also Nansat.reproject()
        skip_gcps : int
            Factor to reduce the number of GCPs by and increase speed
            See also Nansat.reproject()

        Returns
        --------
        watermask : Nansat object with water mask in current projection

        See Also
        ---------
        `<http://www.glcf.umd.edu/data/watermask/>`_

        `<http://nansat.readthedocs.io/en/latest/source/features.html#differentiating-between-land-and-water>`_

        """
        mod44DataExist = True
        # check if path is given in input param or in environment
        if mod44path is None:
            mod44path = os.getenv('MOD44WPATH')
        if mod44path is None:
            mod44DataExist = False
        # check if VRT file exist
        elif not os.path.exists(mod44path + '/MOD44W.vrt'):
            mod44DataExist = False
        self.logger.debug('MODPATH: %s' % mod44path)

        if not mod44DataExist:
            raise IOError('250 meters resolution watermask from MODIS '
                    '44W Product does not exist - see Nansat '
                    'documentation to get it (the path is % s)' % mod44path)

        # MOD44W data does exist: open the VRT file in Nansat
        watermask = Nansat(mod44path + '/MOD44W.vrt', mapper='MOD44W',
                           log_level=self.logger.level)
        # reproject on self or given Domain
        if dst_domain is None:
            dst_domain = self
        lon, lat = dst_domain.get_border()
        watermask.crop_lonlat([lon.min(), lon.max()], [lat.min(), lat.max()])
        watermask.reproject(dst_domain, addmask=False, **kwargs)

        return watermask

    def write_figure(self, filename='', bands=1, clim=None, addDate=False,
                     array_modfunc=None, **kwargs):
        """Save a raster band to a figure in graphical format.

        Get numpy array from the band(s) and band information specified
        either by given band number or band id.
        -- If three bands are given, merge them and create PIL image.
        -- If one band is given, create indexed image
        Create Figure object and:
        Adjust the array brightness and contrast using the given min/max or
        histogram.
        Apply logarithmic scaling of color tone.
        Generate and append legend.
        Save the PIL output image in PNG or any other graphical format.
        If the filename extension is 'tif', the figure file is converted
        to GeoTiff

        Parameters
        -----------
        filename : str
            Output file name. if one of extensions 'png', 'PNG', 'tif',
            'TIF', 'bmp', 'BMP', 'jpg', 'JPG', 'jpeg', 'JPEG' is included,
            specified file is created. otherwise, 'png' file is created.
        bands : integer or string or list (elements are integer or string),
            default = 1
            the size of the list has to be 1 or 3.
            if the size is 3, RGB image is created based on the three bands.
            Then the first element is Red, the second is Green,
            and the third is Blue.
        clim : list with two elements or 'hist' to specify range of colormap
            None (default) : min/max values are fetched from WKV,
            fallback-'hist'
            [min, max] : min and max are numbers, or
            [[min, min, min], [max, max, max]]: three bands used
            'hist' : a histogram is used to calculate min and max values
        addDate : boolean
            False (default) : no date will be aded to the caption
            True : the first time of the object will be added to the caption
        array_modfunc : None
            None (default) : figure created using array in provided band
            function : figure created using array modified by provided function
        **kwargs : parameters for Figure().

        Notes
        ---------
        if filename is specified, creates image file

        Returns
        -------
        Figure : Figure object
            filename extension define format (default format is png)

        Example
        --------
            >>> n.write_figure('test.jpg') # write indexed image
            >>> n.write_figure('test_rgb_hist.jpg', clim='hist', bands=[1, 2, 3]) # RGB image
            >>> n.write_figure('r09_log3_leg.jpg', logarithm=True, legend=True,
                                gamma=3, titleString='Title', fontSize=30,
                                numOfTicks=15) # add legend
            >>> n.write_figure(filename='transparent.png', bands=[3],
                               mask_array=wmArray,
                               mask_lut={0: [0,0,0]},
                               clim=[0,0.15], cmapName='gray',
                               transparency=[0,0,0]) # write transparent image
        See also
        --------
        Figure()

        `<http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps>`_

        """
        # convert <bands> from integer, or string, or list of strings
        # into list of integers
        if isinstance(bands, list):
            for i, band in enumerate(bands):
                bands[i] = self.get_band_number(band)
        else:
            bands = [self.get_band_number(bands)]

        # == create 3D ARRAY ==
        array = None
        for band in bands:
            # get array from band and reshape to (1,height,width)
            iArray = self[band]
            if array_modfunc:
                iArray = array_modfunc(iArray)
            iArray = iArray.reshape(1, iArray.shape[0], iArray.shape[1])
            # create new 3D array or append band
            if array is None:
                array = iArray
            else:
                array = np.append(array, iArray, axis=0)

        # == CREATE FIGURE object and parse input parameters ==
        fig = Figure(array, **kwargs)
        array = None

        # == PREPARE cmin/cmax ==
        # check if cmin and cmax are given as the arguments
        if 'cmin' in kwargs.keys() and 'cmax' in kwargs.keys():
            clim = [kwargs['cmin'], kwargs['cmax']]

        # try to get clim from WKV if it is not given as the argument
        # if failed clim will be evaluated from histogram
        if clim is None:
            clim = [[], []]
            for i, iBand in enumerate(bands):
                try:
                    defValue = (self.vrt.dataset.GetRasterBand(iBand).
                                GetMetadataItem('minmax').split(' '))
                except:
                    clim = 'hist'
                    break
                clim[0].append(float(defValue[0]))
                clim[1].append(float(defValue[1]))

        # Estimate color min/max from histogram
        if clim == 'hist':
            clim = fig.clim_from_histogram(**kwargs)

        # modify clim to the proper shape [[min], [max]]
        # or [[min, min, min], [max, max, max]]
        if (len(clim) == 2 and
           ((isinstance(clim[0], float)) or (isinstance(clim[0], int))) and
           ((isinstance(clim[1], float)) or (isinstance(clim[1], int)))):
            clim = [[clim[0]], [clim[1]]]

        # if the len(clim) is not same as len(bands), the 1st element is used.
        for i in range(2):
            if len(clim[i]) != len(bands):
                clim[i] = [clim[i][0]] * len(bands)

        self.logger.info('clim: %s ' % clim)

        # == PREPARE caption ==
        if 'caption' in kwargs:
            caption = kwargs['caption']
        else:
            # get longName and units from vrt
            band = self.get_GDALRasterBand(bands[0])
            longName = band.GetMetadata().get('long_name', '')
            units = band.GetMetadata().get('units', '')

            # make caption from longname, units
            caption = longName + ' [' + units + ']'

        # add DATE to caption
        if addDate:
            caption += self.time_coverage_start.strftime(' %Y-%m-%d')

        self.logger.info('caption: %s ' % caption)

        # == PROCESS figure ==
        fig.process(cmin=clim[0], cmax=clim[1], caption=caption)

        # == finally SAVE to a image file ==
        fig.save(filename, **kwargs)
        # If tiff image, convert to GeoTiff
        if filename[-3:] == 'tif':
            self.vrt.copyproj(filename)
        return fig


    def write_geotiffimage(self, filename, band_id=1):
        """Writes an 8-bit GeoTiff image for a given band.

        The colormap is fetched from the metadata item 'colormap'. Fallback colormap is 'gray'.

        Color limits are fetched from the metadata item 'minmax'. If 'minmax' is not specified, min
        and max of the raster data is used.

        The method can be replaced by using nansat.write_figure(). However, write_figure uses PIL,
        which does not allow Tiff compression. This gives much larger files.

        Parameters
        -----------
        filename : str
        band_id : int or str

        """
        bandNo = self.get_band_number(band_id)
        band = self.get_GDALRasterBand(band_id)

        minmax = band.GetMetadataItem('minmax')
        # Get min and max from band histogram if not given (from wkv)
        if minmax is None:
            (rmin, rmax) = band.ComputeRasterMinMax()
            minmax = str(rmin) + ' ' + str(rmax)

        bMin = float(minmax.split(' ')[0])
        bMax = float(minmax.split(' ')[1])
        # Make colormap from WKV information
        try:
            colormap = band.GetMetadataItem('colormap')
            cmap = cm.get_cmap(colormap, 256)
            cmap = cmap(np.arange(256)) * 255
        except:
            if not MATPLOTLIB_IS_INSTALLED:
                self.logger.debug('Geotiff is only available in gray '
                                  'since matplotlib is not installed.')
            cmap = np.vstack([np.arange(256.),
                              np.arange(256.),
                              np.arange(256.),
                              np.ones(256)*255]).T
        colorTable = gdal.ColorTable()
        for i in range(cmap.shape[0]):
            colorEntry = (int(cmap[i, 0]), int(cmap[i, 1]),
                          int(cmap[i, 2]), int(cmap[i, 3]))
            colorTable.SetColorEntry(i, colorEntry)
        # Write Tiff image, with data scaled to values between 0 and 255
        outDataset = gdal.GetDriverByName('Gtiff').Create(filename,
                                                          band.XSize,
                                                          band.YSize, 1,
                                                          gdal.GDT_Byte,
                                                          ['COMPRESS=LZW'])
        data = self.__getitem__(bandNo)
        scaledData = ((data - bMin) / (bMax - bMin)) * 255
        outDataset.GetRasterBand(1).WriteArray(scaledData)
        outDataset.GetRasterBand(1).SetMetadata(band.GetMetadata())
        try:
            outDataset.GetRasterBand(1).SetColorTable(colorTable)
        except:
            # Happens after reprojection, a possible bug?
            print('Could not set color table')
            print(colorTable)
        outDataset = None
        self.vrt.copyproj(filename)

    @property
    def time_coverage_start(self):
        return parse_time(self.get_metadata('time_coverage_start'))

    @property
    def time_coverage_end(self):
        return parse_time(self.get_metadata('time_coverage_end'))

    def get_metadata(self, key=None, band_id=None, unescape=True):
        """Get metadata from self.vrt.dataset

        Parameters
        ----------
        key : str
            name of the metadata key. If not givem all metadata is returned
        band_id : int or str
            number or name of band to get metadata from.
            If not given, global metadata is returned
        unescape : bool
            Replace '&quot;', '&amp;', '&lt;' and '&gt;' with these symbols " & < > ?

        Returns
        --------
        metadata : str
            string with metadata if key is given and found
        metadata : dict
            dictionary with all metadata if key is not given

        Raises
        ------
        ValueError, if key is not found

        """
        # get all metadata from dataset or from band
        if band_id is None:
            metadata = self.vrt.dataset.GetMetadata()
        else:
            metadata = self.get_GDALRasterBand(band_id).GetMetadata()

        # remove escapes of special characters
        if unescape:
            for i in metadata:
                metadata[i] = saxutils.unescape(metadata[i], {'&quot;': '"'})

        # get all metadata or from a key
        if key is not None:
            try:
                metadata = metadata[key]
            except KeyError:
                raise ValueError('%s does not have metadata %s' % (self.filename, key))

        return metadata

    def set_metadata(self, key='', value='', band_id=None):
        """Set metadata to self.vrt.dataset

        Parameters
        -----------
        key : string or dictionary with strings
            name of the metadata, or dictionary with metadata names, values
        value : string
            value of metadata
        band_id : int or str
            number or name of band
            Without : global metadata is set

        Notes
        ---------
        self.vrt.dataset : sets metadata in GDAL current dataset

        """
        # set all metadata to the dataset or to the band
        if band_id is None:
            metadata_receiver = self.vrt.dataset
        else:
            bandNumber = self.get_band_number(band_id)
            metadata_receiver = self.vrt.dataset.GetRasterBand(bandNumber)

        # set metadata from dictionary or from single pair key,value
        if type(key) == dict:
            for k in key:
                metadata_receiver.SetMetadataItem(str(k), str(key[k]))
        else:
            metadata_receiver.SetMetadataItem(str(key), str(value))

    def _get_dataset_metadata(self):
        # open GDAL dataset. It will be parsed to all mappers for testing
        gdal_dataset, metadata = None, dict()
        if not self.filename.startswith('http'):
            try:
                gdal_dataset = gdal.Open(self.filename)
            except RuntimeError:
                self.logger.error('GDAL could not open %s, trying to read with Nansat mappers...'
                                  % self.filename)
        if gdal_dataset is not None:
            # get metadata from the GDAL dataset
            metadata = gdal_dataset.GetMetadata()

        return gdal_dataset, metadata


    def _get_mapper(self, mappername, **kwargs):
        """Create VRT file in memory (VSI-file) with variable mapping

        If mappername is given only this mapper will be used,
        else loop over all available mappers in mapperList to get the
        matching one.
        In the loop :
            If the specific error appears the mapper is not used
            and the next mapper is tested.
            Otherwise the mapper returns VRT.
        If type of the sensor is identified, add mapping variables.
        If all mappers fail, make simple copy of the input DS into a VSI/VRT

        Parameters
        -----------
        mappername : string, optional (e.g. 'ASAR' or 'merisL2')

        Returns
        --------
        tmp_vrt : VRT object
            tmp_vrt.dataset is a GDAL VRT dataset

        Raises
        --------
        IOError : occurs if the input file does not exist
        ValueError : occurs if given mapper cannot open the input file
        NansatReadError : occurs if no mapper fits the input file

        """
        if os.path.isfile(self.filename):
            # Make sure file exists and can be opened for reading before proceeding
            assert os.access(self.filename, os.R_OK)
        else:
            for f in glob.glob(os.path.join(self.filename, '*.*')):
                assert os.access(f, os.R_OK)
        # lazy import of nansat mappers
        # if nansat mappers were not imported yet
        global nansatMappers
        if nansatMappers is None:
            nansatMappers = _import_mappers()

        # open GDAL dataset. It will be parsed to all mappers for testing
        gdal_dataset, metadata = self._get_dataset_metadata()
        tmp_vrt = None

        # TODO: There seems to be code repetition in this if-test - should be avoided...
        if mappername is not '':
            # If a specific mapper is requested, we test only this one.
            # get the module name
            mappername = 'mapper_' + mappername.replace('mapper_', '').replace('.py', '').lower()
            # check if the mapper is available
            if mappername not in nansatMappers:
                raise ValueError('Mapper ' + mappername + ' not found')

            # check if mapper is importbale or raise an ImportError error
            if isinstance(nansatMappers[mappername], tuple):
                errType, err, traceback = nansatMappers[mappername]
                # self.logger.error(err, exc_info=(errType, err, traceback))
                # TODO: python 3.6 does not support with syntax
                raise EnvironmentError
                #raise errType, err, traceback

            # create VRT using the selected mapper
            tmp_vrt = nansatMappers[mappername](self.filename, gdal_dataset, metadata, **kwargs)
            self.mapper = mappername.replace('mapper_', '')
        else:
            # We test all mappers, import one by one
            import_errors = []
            for iMapper in nansatMappers:
                # skip non-importable mappers
                if isinstance(nansatMappers[iMapper], tuple):
                    # keep errors to show before use of generic mapper
                    import_errors.append(nansatMappers[iMapper][1])
                    continue

                self.logger.debug('Trying %s...' % iMapper)

                # show all ImportError warnings before trying generic_mapper
                if iMapper == 'mapper_generic' and len(import_errors) > 0:
                    self.logger.error('\nWarning! The following mappers failed:')
                    for ie in import_errors:
                        self.logger.error(import_errors)

                # create a Mapper object and get VRT dataset from it
                try:
                    tmp_vrt = nansatMappers[iMapper](self.filename, gdal_dataset, metadata, **kwargs)
                    self.logger.info('Mapper %s - success!' % iMapper)
                    self.mapper = iMapper.replace('mapper_', '')
                    break
                except WrongMapperError:
                    pass

        # if no mapper fits, make simple copy of the input DS into a VSI/VRT
        if tmp_vrt is None and gdal_dataset is not None:
            self.logger.warning('No mapper fits, returning GDAL bands!')
            tmp_vrt = VRT.from_gdal_dataset(gdal_dataset, metadata=metadata)
            for iBand in range(gdal_dataset.RasterCount):
                tmp_vrt.create_band({'SourceFilename': self.filename,
                                     'SourceBand': iBand + 1})
                tmp_vrt.dataset.FlushCache()
            self.mapper = 'gdal_bands'

        # if GDAL cannot open the file, and no mappers exist which can make VRT
        if tmp_vrt is None and gdal_dataset is None:
            # check if given data file exists
            if not os.path.isfile(self.filename):
                raise IOError('%s: File does not exist' % (self.filename))
            raise NansatReadError('%s: File cannot be read with NANSAT - '
                    'consider writing a mapper' % self.filename)

        return tmp_vrt

    def get_band_number(self, band_id):
        """Return absolute band number

        Check if given band_id is valid
        Return absolute number of the band in the VRT

        Parameters
        ----------
        band_id : int or str or dict
            - if int : checks if such band exists and returns band_id
            - if str : finds band with coresponding name
            - if dict : finds first band with given metadata

        Returns
        --------
        absolute band number : int

        """
        band_number = 0
        # if band_id is str: create simple dict with seraching criteria
        if type(band_id) == str:
            band_id = {'name': band_id}

        # if band_id is dict: search self.bands with seraching criteria
        if type(band_id) == dict:
            bands_meta = self.bands()
            for b in bands_meta:
                num_correct_keys = 0
                for key in band_id:
                    if (key in bands_meta[b] and
                            band_id[key] == bands_meta[b][key]):
                        num_correct_keys = num_correct_keys + 1
                    if num_correct_keys == len(band_id):
                        band_number = b
                        break

        # if band_id is int and with bounds: return this number
        if (type(band_id) == int and band_id >= 1 and
                band_id <= self.vrt.dataset.RasterCount):
            band_number = band_id

        # if no band_number found - raise error
        if band_number == 0:
            raise ValueError('Cannot find band %s! '
                              'band_number is from 1 to %s'
                              % (str(band_id), self.vrt.dataset.RasterCount))

        return band_number

    def get_transect(self, points, bands,
                        lonlat=True,
                        smooth_radius=0,
                        smooth_function=nanmedian,
                        data=None,
                        cornersonly=False):
        """Get values from transect from given vector of poins

        Parameters
        ----------
        points : 2xN list or array, N (number of points) >= 1
            coordinates [[x1, x2, y2], [y1, y2, y3]]
        bands : list of int or string
            elements of the list are band number or band Name
        lonlat : bool
            If the points in lat/lon, then True.
            If the points in pixel/line, then False.
        smooth_radius: int
            If smootRadius is greater than 0, smooth every transect
            pixel as the median or mean value in a circule with radius
            equal to the given number.
        smooth_function: func
            function for averaging values collected within smooth radius
        data : ndarray
            alternative array with data to take values from

        Returns
        --------
        transect : numpy record array

        """
        # check if points is 2D array with shape 2xN (N>=1)
        if (len(np.shape(points)) != 2 or
              np.shape(points)[0] != 2 or
              np.shape(points)[1] < 1):
            # points are not 2xN array
            raise ValueError('Input points must be 2xN array with N>0')

        # get names of bands
        band_names = []
        for band in bands:
            try:
                bandN = self.get_band_number(band)
            except ValueError:
                self.logger.error('Wrong band name %s' % band)
            else:
                band_names.append(self.bands()[bandN]['name'])

        if data is not None:
            band_names.append('input')

        pixVector, linVector = self._get_pix_lin_vectors(points, lonlat, cornersonly, smooth_radius)

        # create output transect
        t = np.recarray((len(pixVector)), dtype=[('pixel', int),
                                                ('line', int),
                                                ('lon', float),
                                                ('lat', float), ])

        # add pixel, line, lon, lat values to output
        t['pixel'] = pixVector
        t['line'] = linVector
        t['lon'], t['lat'] = self.transform_points(t['pixel'], t['line'], DstToSrc=0)

        # get values from bands or input data
        for band_name in band_names:
            t = self._extract_transect_data(t, band_name, data, smooth_radius, smooth_function)

        return t

    def _extract_transect_data(self, t, band_name, data, smooth_radius, smooth_function):
        """Extract data along transect from input band"""
        # mask for extraction within circular area
        xgrid, ygrid = np.mgrid[0:smooth_radius * 2 + 1, 0:smooth_radius * 2 + 1]
        distance = ((xgrid - smooth_radius) ** 2 + (ygrid - smooth_radius) ** 2) ** 0.5
        mask = distance <= smooth_radius

        if band_name == 'input':
            bandArray = data
        else:
            bandArray = self[band_name]
        # average values from pixel inside a circle
        bandValues = []
        for r, c in zip(t['line'], t['pixel']):
            subarray = bandArray[r-smooth_radius:r+smooth_radius+1,
                                 c-smooth_radius:c+smooth_radius+1]
            bandValues.append(smooth_function(subarray[mask]))
        return append_fields(t, band_name, bandValues).data

    def digitize_points(self, band=1, **kwargs):
        """Get coordinates of interactively digitized points

        Parameters
        ----------
        band : int or str
            ID of Nansat band
        **kwargs : keyword arguments for imshow

        Returns
        --------
        points : list
            list of 2xN arrays of points to be used in Nansat.get_transect()

        """
        data = self[band]
        browser = PointBrowser(data, **kwargs)
        points = browser.get_points()

        return points

    def crop_interactive(self, band=1, maxwidth=1000, **kwargs):
        """Interactively select boundary and crop Nansat object

        Parameters
        ----------
        band : int or str
            id of the band to show for interactive selection of boundaries
        maxwidth : int
            large input data is downscaled to <maxwidth>
        **kwargs : keyword arguments for imshow

        Notes
        --------
        self.vrt : VRT
            superVRT is created with modified SrcRect and DstRect

        Returns
        -------
        extent : (x_offset, y_offset, x_size, y_size)
            x_offset  - X offset in the original dataset
            y_offset  - Y offset in the original dataset
            x_size - width of the new dataset
            y_size - height of the new dataset

        Examples
        --------
            >>> extent = n.crop_interactive(band=1) # crop a subimage interactively

        """
        resized = False
        if self.shape()[1] > maxwidth:
            factor = self.resize(width=1000)
            resized = True
        else:
            factor = 1
        # use interactive PointBrowser for selecting extent
        try:
           points = self.digitize_points(band=band,**kwargs)[0]
        except:
           if resized:
              self.undo()
           return

        x_offset, x_size = Nansat._get_crop_offset_size(0, points, factor)
        y_offset, y_size = Nansat._get_crop_offset_size(1, points, factor)

        if resized:
            self.undo()

        return self.crop(x_offset, y_offset, x_size, y_size)

    def crop_lonlat(self, lonlim, latlim):
        """Crop Nansat object to fit into given longitude/latitude limit

        Parameters
        ----------
        lonlim : list of 2 float
            min/max of longitude
        latlim : list of 2 float
            min/max of latitude

        Notes
        --------
        self.vrt : VRT
            crops vrt to size that corresponds to lon/lat limits

        Returns
        -------
        extent : (x_offset, y_offset, x_size, y_size)
            x_offset  - X offset in the original dataset
            y_offset  - Y offset in the original dataset
            x_size - width of the new dataset
            y_size - height of the new dataset

        Examples
        --------
            >>> extent = n.crop(lonlim=[-10,10], latlim=[-20,20]) # crop for given lon/lat limits

        """
        # lon/lat lists for four corners
        lon_corners = [lonlim[0], lonlim[0], lonlim[1], lonlim[1]]
        lat_corners = [latlim[0], latlim[1], latlim[0], latlim[1]]
        pix_corners, lin_corners = self.transform_points(lon_corners, lat_corners, 1)

        offset_size = lambda corners: (round(min(corners)), round(max(corners) - min(corners)))
        x_offset, x_size = offset_size(pix_corners)
        y_offset, y_size = offset_size(lin_corners)

        return self.crop(x_offset, y_offset, x_size, y_size)

    @staticmethod
    def _get_crop_offset_size(axis, points, factor):
        """Get offset and size of cropped image"""
        offset = round(points.min(axis=1)[axis] / factor)
        size = round((points.max(axis=1)[axis] - offset) / factor)
        return offset, size

    @staticmethod
    def _fix_crop_offset_size(offset, size, raster_size, allow_larger=False):
        """Check and correct default values of invalid offset or size"""
        # test if crop is totally outside
        if (offset > raster_size or (offset + size) < 0):
            raise ValueError('Cropping region is outside the image! offset: %f size: %f'
                               %(float(offset), float(size)))
        if allow_larger:
            return offset, size

        if offset < 0:
            size += offset
            offset = 0
        if (size + offset) > raster_size:
            size = raster_size - offset
        return offset, size

    def crop(self, x_offset, y_offset, x_size, y_size, allow_larger=False):
        """Crop Nansat object

        Create superVRT, modify the Source Rectangle (SrcRect) and Destination
        Rectangle (DstRect) tags in the VRT file for each band in order
        to take only part of the original image,
        create new GCPs or new GeoTransform for the cropped object.

        Parameters
        ----------
        x_offset : int
            pixel offset of subimage
        y_offset : int
            line offset of subimage
        x_size : int
            width in pixels of subimage
        y_size : int
            height in pizels of subimage
        allow_larger : bool
            Allow resulting extent to be larger than the original image?

        Notes
        --------
        self.vrt : super-VRT is created with modified SrcRect and DstRect

        Returns
        -------
        extent : (x_offset, y_offset, x_size, y_size)
            x_offset  - X offset in the original dataset
            y_offset  - Y offset in the original dataset
            x_size - width of the new dataset
            y_size - height of the new dataset

        Examples
        --------
            >>> extent = n.crop(10, 20, 100, 200)

        """
        x_offset, x_size = Nansat._fix_crop_offset_size(x_offset, x_size,
                                                        self.shape()[1], allow_larger)
        y_offset, y_size = Nansat._fix_crop_offset_size(y_offset, y_size,
                                                        self.shape()[0], allow_larger)

        extent = (int(x_offset), int(y_offset), int(x_size), int(y_size))
        self.logger.debug('x_offset: %d, y_offset: %d, x_size: %d, y_size: %d' % extent)

        # test if crop is larger or equal to image size
        if x_offset == y_offset == 0 and (y_size, x_size) == self.shape():
            self.logger.error(('WARNING! Cropping region is larger or equal to image!'))
            return extent

        # create super VRT and change it
        self.vrt = self.vrt.get_super_vrt()
        self.vrt.set_offset_size('x', x_offset, x_size)
        self.vrt.set_offset_size('y', y_offset, y_size)
        self.vrt.shift_cropped_gcps(x_offset, x_size, y_offset, y_size)
        self.vrt.shift_cropped_geo_transform(x_offset, x_size, y_offset, y_size)
        return extent

    def extend(self, left=0, right=0, top=0, bottom=0):
        """Extend domain from four sides

        Parameters
        ----------
        left : int
            number of pixels to add from left side
        right : int
            number of pixels to add from right side
        top : int
            number of pixels to add from top side
        bottom : int
            number of pixels to add from bottom side

        Notes
        -----
        Canges self.vrt by adding nexgative offset or setting size to be large that original size.

        """
        x_offset, y_offset = -left, -top
        x_size = self.shape()[1] + left + right
        y_size = self.shape()[0] + top + bottom
        self.crop(x_offset, y_offset, x_size, y_size, allow_larger=True)

    def _get_pix_lin_vectors(self, points, lonlat, cornersonly, smooth_radius):
        """Get vectors with pixel and line values for input corner points"""
        # if points in degree, convert them into pix/lin
        if lonlat:
            pix, lin = self.transform_points(points[0], points[1], DstToSrc=1)
        else:
            pix, lin = points[0], points[1]

        if cornersonly:
            return pix, lin

        # full vectors of pixel coordinates based on coordinates of vertices
        pixVector, linVector = [pix[0]], [lin[0]]
        for pn in range(len(pix[1:])):
            px0, px1 = pix[pn], pix[pn+1]
            py0, py1 = lin[pn], lin[pn+1]
            length = int(np.round(np.hypot(px1-px0, py0-py1)))
            pixVector += list(np.linspace(px0, px1, length+1)[1:])
            linVector += list(np.linspace(py0, py1, length+1)[1:])

        # remove out of region points
        pixVector = np.floor(pixVector)
        linVector = np.floor(linVector)
        gpi = ((pixVector >= (0 + smooth_radius)) *
               (linVector >= (0 + smooth_radius)) *
               (pixVector < (self.shape()[1] - smooth_radius)) *
               (linVector < (self.shape()[0] - smooth_radius)))

        return pixVector[gpi], linVector[gpi]


def _import_mappers(log_level=None):
    """Import available mappers into a dictionary

    Returns
    --------
    nansat_mappers : dict
        key  : mapper name
        value: class Mapper(VRT) from the mappers module

    """
    logger = add_logger('import_mappers', logLevel=log_level)
    # import built-in mappers
    import nansat.mappers
    mapper_packages = [nansat.mappers]

    # import user-defined mappers (if any)
    try:
        import nansat_mappers as nansat_mappers_pkg
    except ImportError:
        pass
    else:
        logger.info('User defined mappers found in %s' % nansat_mappers_pkg.__path__)
        mapper_packages = [nansat_mappers_pkg, nansat.mappers]

    # create ordered dict for mappers
    nansat_mappers = OrderedDict()
    for mapper_package in mapper_packages:
        logger.debug('From package: %s' % mapper_package.__path__)
        # scan through modules and load all modules that contain class Mapper
        for finder, name, ispkg in (pkgutil.iter_modules(mapper_package.__path__)):
            logger.debug('Loading mapper %s' % name)
            # Only mappers containing 'mapper' in the module name should be returned
            if not 'mapper' in name: continue
            loader = finder.find_module(name)
            # try to import mapper module
            module = loader.load_module(name)
            # add the imported mapper to nansat_mappers
            if hasattr(module, 'Mapper'):
                nansat_mappers[name] = module.Mapper

        # move netcdfcdf mapper to the end
        if 'mapper_netcdf_cf' in nansat_mappers:
            nansat_mappers['mapper_netcdf_cf'] = nansat_mappers.pop('mapper_netcdf_cf')

        # move generic_mapper to the end
        if 'mapper_generic' in nansat_mappers:
            nansat_mappers['mapper_generic'] = nansat_mappers.pop('mapper_generic')

    return nansat_mappers



#------------------------------------------------------------------------------
# Name:         test_nansat.py
# Purpose:      Test the nansat module
#
# Author:       Morten Wergeland Hansen, Anton Korosov, Asuka Yamakawa
# Modified: Morten Wergeland Hansen
#
# Created:  18.06.2014
# Last modified:27.08.2014 10:58
# Copyright:    (c) NERSC
# License:
#------------------------------------------------------------------------------
import unittest
from nansat import NSR
from nansat.utils import osr

from nansat.exceptions import NansatProjectionError


class NSRTest(unittest.TestCase):
    def test_init_empty(self):
        nsr = NSR()

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 0)

    def test_init_from_none(self):
        nsr = NSR(None)

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 5)

    def test_init_from_0(self):
        nsr = NSR(0)

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 0)

    def test_init_from_EPSG(self):
        nsr = NSR(4326)

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 0)
        self.assertTrue('4326' in nsr.ExportToWkt())

    def test_init_from_proj4(self):
        nsr = NSR('+proj=longlat')

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 0)
        self.assertTrue('longlat' in nsr.ExportToProj4())

    def test_init_from_proj4_unicode(self):
        nsr = NSR(u'+proj=longlat')

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 0)
        self.assertTrue('longlat' in nsr.ExportToProj4())

    def test_init_from_wkt(self):
        nsr = NSR(
        'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,'\
        'AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,'\
        'AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],'\
        'AXIS["Latitude",NORTH],AXIS["Longitude",EAST],AUTHORITY["EPSG","4326"]]')

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 0)
        self.assertTrue('longlat' in nsr.ExportToProj4())

    def test_init_from_NSR(self):
        nsr = NSR(NSR(4326))

        self.assertEqual(type(nsr), NSR)
        self.assertEqual(nsr.Validate(), 0)
        self.assertTrue('longlat' in nsr.ExportToProj4())

    def test_dont_init_from_invalid(self):
        self.assertRaises(NansatProjectionError, NSR, -10)
        self.assertRaises(NansatProjectionError, NSR, 'some crap')
        ss = osr.SpatialReference()
        self.assertRaises(NansatProjectionError, NSR, ss)
